# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In-silico mutagenesis (ISM) functions for sequence interpretation."""

from collections.abc import Sequence

from alphagenome.data import genome
import numpy as np


def ism_variants(
    interval: genome.Interval,
    sequence: str,
    vocabulary: str = 'ACGT',
    skip_n: bool = False,
) -> list[genome.Variant]:
  """Create a list of all possible single nucleotide variants for an interval.

  Args:
    interval: Interval for which to generate the variants.
    sequence: Sequence extracted from the reference genome at the `interval`.
      Needs to have the same length as `interval.width`.
    vocabulary: Vocabulary of possible alternative bases contained in
      `sequence`.
    skip_n: If True, skip the N bases.

  Returns:
    List of all possible single nucleotide variants for a genomic interval.
  """
  if len(sequence) != interval.width:
    raise ValueError('Sequence must have the same length as interval.')
  variants = []
  for position in range(interval.width):
    reference_base = sequence[position]
    if skip_n and reference_base == 'N':
      continue
    for alternative_base in vocabulary:
      if reference_base == alternative_base:
        continue
      variants.append(
          genome.Variant(
              chromosome=interval.chromosome,
              reference_bases=reference_base,
              alternate_bases=alternative_base,
              position=interval.start + position + 1,
          )
      )
  return variants


def ism_matrix(
    variant_scores: Sequence[float],
    variants: Sequence[genome.Variant],
    interval: genome.Interval | None = None,
    multiply_by_sequence: bool = True,
    vocabulary: str = 'ACGT',
    require_fully_filled: bool = True,
) -> np.ndarray:
  """Construct the ISM (position, base) matrix from individual ISM scores.

  This function returns the relative effect of the variants compared to the
  per-position average: score[position, base] - mean(score[position, :]]).

  Args:
    variant_scores: Variant effect scores corresponding to the variants. These
      could be obtained from the score_variants() output summarised to a single
      scalar. Summarisation could be for example be obtained by selecting a
      specific variant scorer output and extract a specific value from the
      (variant, track) matrix.
    variants: Sequence of variants used to transform into the ISM matrix.
    interval: Interval for which to get the contribution scores. All variants
      need to be contained within that interval. If None, it will be
      automatically inferred from variants.
    multiply_by_sequence: If True, only return non-zero values at
      one-hot-encoded reference genome sequence bases.
    vocabulary: Vocabulary of possible alternative bases contained in
      `sequence`. The order determines the column order of the returned matrix.
    require_fully_filled: If True, raise an error if not all positions are
      covered by variants.

  Returns:
    Matrix of shape (interval.width, 4) containing variant scores.
  """
  if len(variants) != len(variant_scores):
    raise ValueError(
        'Variants and variant_scores need to have the same length.'
    )
  if interval is None:
    interval = genome.Interval(
        chromosome=variants[0].chromosome,
        start=min(variant.start for variant in variants),
        end=max(variant.end for variant in variants),
    )
  scores = np.zeros((interval.width, 4), dtype=np.float32)
  filled = np.zeros((interval.width, 4), dtype=bool)
  base_index = {base: i for i, base in enumerate(vocabulary)}

  for variant, score in zip(variants, variant_scores, strict=True):
    if len(variant.alternate_bases) != 1 or len(variant.reference_bases) != 1:
      # Only looking for single nucleotide variants.
      continue
    if not interval.contains(variant.reference_interval):
      continue
    position = variant.start - interval.start
    scores[position, base_index[variant.alternate_bases]] = score
    filled[position, base_index[variant.alternate_bases]] = True

  # Check that all positions were covered by variants.
  if (filled.sum(axis=-1) == 0).any() and require_fully_filled:
    missing_positions = list(
        np.where(filled.sum(axis=-1) == 0)[0] + interval.start + 1
    )
    raise ValueError(
        'No variants were found for the (1-based) positions on chromosome '
        f'{interval.chromosome}: {missing_positions}'
    )
  elif (filled.sum(axis=-1) == 4).any():
    full_positions = list(
        np.where(filled.sum(axis=-1) == 4)[0] + interval.start + 1
    )
    raise ValueError(
        'Some variants were found with 4 different alternative bases for the'
        ' position instead of 3. List of positions on chromosome '
        f'{interval.chromosome}: {full_positions}'
    )
  elif (filled.sum(axis=-1) != 3).any() and require_fully_filled:
    missing_positions = list(
        np.where(filled.sum(axis=-1) != 3)[0] + interval.start + 1
    )
    raise ValueError(
        'Some variants were found with only 1 or 2 alternative bases for the'
        ' position instead of 3. List of positions on chromosome '
        f'{interval.chromosome}: {missing_positions}'
    )

  scores -= np.sum(scores, axis=-1, keepdims=True) / (len(vocabulary) - 1)
  if multiply_by_sequence:
    scores = scores * (~filled)
  return scores



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hatch build hook to generate Python bindings for protos."""

import os
from typing import Any
from grpc_tools import protoc
from hatchling.builders.hooks.plugin.interface import BuildHookInterface  # pylint: disable=g-importing-member


_ROOT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'src')

# Tuple of proto message definitions to build Python bindings for. Paths must
# be relative to root directory.
_ALPHAGENOME_PROTOS = (
    'alphagenome/protos/dna_model.proto',
    'alphagenome/protos/dna_model_service.proto',
    'alphagenome/protos/tensor.proto',
)


class GenerateProtos(BuildHookInterface):
  """Generates Python protobuf bindings for alphagenome.protos."""

  def initialize(self, version: str, build_data: dict[str, Any]) -> None:
    del version, build_data  # Unused.

    for proto_path in _ALPHAGENOME_PROTOS:
      proto_args = [
          'grpc_tools.protoc',
          f'--proto_path={_ROOT_DIR}',
          f'--python_out={_ROOT_DIR}',
          f'--grpc_python_out={_ROOT_DIR}',
          os.path.join(_ROOT_DIR, proto_path),
      ]
      if protoc.main(proto_args) != 0:
        raise RuntimeError(f'ERROR: {proto_args}')



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
import dataclasses
import functools
import time
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome import tensor_utils
from alphagenome.data import genome
from alphagenome.data import junction_data
from alphagenome.data import track_data
from alphagenome.interpretation import ism
from alphagenome.models import dna_client
from alphagenome.models import interval_scorers
from alphagenome.models import junction_data_utils
from alphagenome.models import track_data_utils
from alphagenome.models import variant_scorers
from alphagenome.protos import dna_model_pb2
from alphagenome.protos import dna_model_service_pb2
from alphagenome.protos import dna_model_service_pb2_grpc
import anndata
import grpc
import numpy as np
import pandas as pd


class _FakeRpcError(grpc.RpcError):
  """Fake RPC error for testing."""

  def __init__(self, code: grpc.StatusCode):
    self._code = code

  def code(self) -> grpc.StatusCode:
    return self._code


def _generate_prediction_protos(
    predictions: dna_client.Output,
    requested_outputs: Sequence[dna_model_pb2.OutputType],
    bytes_per_chunk: int,
    response_proto_type: ...,
):
  """Helper to generate proto responses for the provided predictions."""
  for output_type in requested_outputs:
    prediction = predictions.get(dna_client.OutputType(output_type))
    if prediction is not None:
      if isinstance(prediction, track_data.TrackData):
        proto, chunks = track_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield response_proto_type(
            output=dna_model_pb2.Output(
                output_type=output_type, track_data=proto
            )
        )
      elif isinstance(prediction, junction_data.JunctionData):
        proto, chunks = junction_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield response_proto_type(
            output=dna_model_pb2.Output(
                output_type=output_type, junction_data=proto
            )
        )
      else:
        raise ValueError(f'Unsupported prediction type: {type(prediction)}')

      for chunk in chunks:
        yield response_proto_type(tensor_chunk=chunk)


def _generate_variant_protos(
    predictions: dna_client.VariantOutput,
    requested_outputs: Sequence[dna_model_pb2.OutputType],
    bytes_per_chunk: int,
):
  """Helper to generate variant proto responses for the provided predictions."""
  for output_type in requested_outputs:
    prediction = predictions.reference.get(dna_client.OutputType(output_type))
    if prediction is not None:
      if isinstance(prediction, track_data.TrackData):
        proto, chunks = track_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield dna_model_service_pb2.PredictVariantResponse(
            reference_output=dna_model_pb2.Output(
                output_type=output_type, track_data=proto
            )
        )
      elif isinstance(prediction, junction_data.JunctionData):
        proto, chunks = junction_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield dna_model_service_pb2.PredictVariantResponse(
            reference_output=dna_model_pb2.Output(
                output_type=output_type, junction_data=proto
            )
        )
      else:
        raise ValueError(f'Unsupported prediction type: {type(prediction)}')

      for chunk in chunks:
        yield dna_model_service_pb2.PredictVariantResponse(tensor_chunk=chunk)

  for output_type in requested_outputs:
    prediction = predictions.alternate.get(dna_client.OutputType(output_type))
    if prediction is not None:
      if isinstance(prediction, track_data.TrackData):
        proto, chunks = track_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield dna_model_service_pb2.PredictVariantResponse(
            alternate_output=dna_model_pb2.Output(
                output_type=output_type, track_data=proto
            )
        )
      elif isinstance(prediction, junction_data.JunctionData):
        proto, chunks = junction_data_utils.to_protos(
            prediction, bytes_per_chunk=bytes_per_chunk
        )
        yield dna_model_service_pb2.PredictVariantResponse(
            alternate_output=dna_model_pb2.Output(
                output_type=output_type, junction_data=proto
            )
        )
      else:
        raise ValueError(f'Unsupported prediction type: {type(prediction)}')

      for chunk in chunks:
        yield dna_model_service_pb2.PredictVariantResponse(tensor_chunk=chunk)


def _generate_variant_scoring_protos(
    scores: list[anndata.AnnData],
    bytes_per_chunk: int,
    response_proto_type: ...,
    variant: genome.Variant | None = None,
):
  """Helper to generate variant scoring responses for the provided scores."""
  for score in scores:
    track_metadata_protos = [
        dna_model_pb2.TrackMetadata(
            name=name,
            strand=genome.Strand.from_str(strand).to_proto(),
        )
        for name, strand in score.var[['name', 'strand']].values
    ]
    gene_metadata_protos = []
    if score.obs is not None and 'gene_id' in score.obs:
      for _, row in score.obs.iterrows():
        if (strand := row.get('strand')) is not None:
          strand = genome.Strand.from_str(strand).to_proto()
        gene_metadata_protos.append(
            dna_model_pb2.GeneScorerMetadata(
                gene_id=row['gene_id'],
                strand=strand,
                name=row.get('gene_name'),
                type=row.get('gene_type'),
                junction_start=row.get('junction_Start'),
                junction_end=row.get('junction_End'),
            )
        )

    if 'quantiles' in score.layers:
      score_tensor = np.stack([score.X, score.layers['quantiles']])
    else:
      score_tensor = score.X[np.newaxis]

    tensor_proto, chunks = tensor_utils.pack_tensor(
        score_tensor, bytes_per_chunk=bytes_per_chunk
    )
    yield response_proto_type(
        output=dna_model_pb2.ScoreVariantOutput(
            variant_data=dna_model_pb2.VariantData(
                values=tensor_proto,
                metadata=dna_model_pb2.VariantMetadata(
                    variant=variant.to_proto() if variant else None,
                    track_metadata=track_metadata_protos,
                    gene_metadata=gene_metadata_protos,
                ),
            )
        )
    )
    for chunk in chunks:
      yield response_proto_type(tensor_chunk=chunk)


def _generate_interval_scoring_protos(
    scores: list[anndata.AnnData],
    bytes_per_chunk: int,
    response_proto_type: ...,
    interval: genome.Interval,
):
  """Helper to generate interval scoring responses for the provided scores."""
  for score in scores:
    track_metadata_protos = [
        dna_model_pb2.TrackMetadata(
            name=name,
            strand=genome.Strand.from_str(strand).to_proto(),
        )
        for name, strand in score.var[['name', 'strand']].values
    ]
    gene_metadata_protos = []
    if score.obs is not None and 'gene_id' in score.obs:
      for _, row in score.obs.iterrows():
        if (strand := row.get('strand')) is not None:
          strand = genome.Strand.from_str(strand).to_proto()
        gene_metadata_protos.append(
            dna_model_pb2.GeneScorerMetadata(
                gene_id=row['gene_id'],
                strand=strand,
                name=row.get('gene_name'),
                type=row.get('gene_type'),
            )
        )

    tensor_proto, chunks = tensor_utils.pack_tensor(
        score.X[np.newaxis], bytes_per_chunk=bytes_per_chunk
    )
    yield response_proto_type(
        output=dna_model_pb2.ScoreIntervalOutput(
            interval_data=dna_model_pb2.IntervalData(
                values=tensor_proto,
                metadata=dna_model_pb2.IntervalMetadata(
                    interval=interval.to_proto(),
                    track_metadata=track_metadata_protos,
                    gene_metadata=gene_metadata_protos,
                ),
            )
        )
    )
    for chunk in chunks:
      yield response_proto_type(tensor_chunk=chunk)


def _create_example_model_predictions(
    *,
    sequence_length: int = 0,
    value: float = 0.0,
    interval: genome.Interval | None = None,
) -> dna_client.Output:
  """Helper to generate example model predictions."""
  if sequence_length == 0:
    if interval is None:
      raise ValueError('Interval must be provided for zero length sequences.')
    sequence_length = interval.width

  return dna_client.Output(
      atac=track_data.TrackData(
          values=np.full((sequence_length, 32), value, dtype=np.float32),
          metadata=pd.DataFrame({
              'name': [f'{i}' for i in range(32)],
              'strand': ['.'] * 32,
          }),
          interval=interval,
      ),
      rna_seq=track_data.TrackData(
          values=np.full((sequence_length, 64), value, dtype=np.float32),
          metadata=pd.DataFrame({
              'name': [f'{i}' for i in range(64)],
              'strand': ['.'] * 64,
          }),
          interval=interval,
      ),
      splice_junctions=junction_data.JunctionData(
          junctions=np.array([
              genome.Interval('chr1', 0, 1000, strand='+'),
              genome.Interval('chr1', 1500, 2000, strand='-'),
          ]),
          values=np.full([2, 3], value, dtype=np.float32),
          metadata=pd.DataFrame({'name': [f'{i}' for i in range(3)]}),
          interval=interval,
      ),
  )


def _create_mock_channel_and_stream(
    response_callable, *, bidi_stream: bool = True
) -> ...:
  mock_stream = mock.create_autospec(
      grpc.StreamStreamMultiCallable, side_effect=response_callable
  )
  mock_channel = mock.create_autospec(grpc.Channel)
  if bidi_stream:
    mock_channel.stream_stream.return_value = mock_stream
  else:
    mock_channel.unary_stream.return_value = mock_stream
  return mock_channel, mock_stream


class ClientTest(parameterized.TestCase):

  def _assert_output_equal(
      self,
      actual: dna_client.Output,
      expected: dna_client.Output,
      msg: ...,
  ) -> None:
    """Helper function to compare Output objects."""
    self.assertEqual(actual.atac, expected.atac, msg)
    self.assertEqual(actual.cage, expected.cage, msg)
    self.assertEqual(actual.dnase, expected.dnase, msg)
    self.assertEqual(actual.rna_seq, expected.rna_seq, msg)
    self.assertEqual(actual.chip_histone, expected.chip_histone, msg)
    self.assertEqual(actual.chip_tf, expected.chip_tf, msg)

    self.assertEqual(actual.splice_sites, expected.splice_sites, msg)
    self.assertEqual(actual.splice_site_usage, expected.splice_site_usage, msg)
    self.assertEqual(actual.splice_junctions, expected.splice_junctions)
    self.assertEqual(actual.contact_maps, expected.contact_maps, msg)

  def _assert_track_metadata_equal(
      self,
      actual: track_data.TrackMetadata,
      expected: track_data.TrackMetadata,
      msg: ...,
  ) -> None:
    if actual is not None and expected is not None:
      pd.testing.assert_frame_equal(actual, expected)
    else:
      self.assertEqual(actual, expected, msg)

  def _assert_output_metadata_equal(
      self,
      actual: dna_client.OutputMetadata,
      expected: dna_client.OutputMetadata,
      msg: ...,
  ) -> None:
    """Helper function to compare OutputMetadata objects."""
    self._assert_track_metadata_equal(actual.atac, expected.atac, msg)
    self._assert_track_metadata_equal(actual.cage, expected.cage, msg)
    self._assert_track_metadata_equal(actual.dnase, expected.dnase, msg)
    self._assert_track_metadata_equal(actual.rna_seq, expected.rna_seq, msg)
    self._assert_track_metadata_equal(
        actual.chip_histone, expected.chip_histone, msg
    )
    self._assert_track_metadata_equal(actual.chip_tf, expected.chip_tf, msg)
    self._assert_track_metadata_equal(
        actual.splice_sites, expected.splice_sites, msg
    )
    self._assert_track_metadata_equal(
        actual.splice_site_usage, expected.splice_site_usage, msg
    )
    self._assert_track_metadata_equal(
        actual.splice_junctions, expected.splice_junctions, msg
    )

  def _assert_track_data_equal(
      self,
      actual: track_data.TrackData,
      expected: track_data.TrackData,
      msg: ...,
  ) -> None:
    """Helper function to compare TrackData objects."""
    pd.testing.assert_frame_equal(actual.metadata, expected.metadata)
    self.assertEqual(actual.interval, expected.interval, msg)
    np.testing.assert_array_equal(actual.values, expected.values)
    self.assertEqual(actual.uns, expected.uns, msg)

  def _assert_junction_data_equal(
      self,
      actual: junction_data.JunctionData,
      expected: junction_data.JunctionData,
      msg: ...,
  ) -> None:
    pd.testing.assert_frame_equal(actual.metadata, expected.metadata)
    np.testing.assert_array_equal(actual.junctions, expected.junctions)
    np.testing.assert_array_equal(actual.values, expected.values)
    self.assertEqual(actual.interval, expected.interval, msg)
    self.assertEqual(actual.uns, expected.uns, msg)

  def _assert_anndata_equal(
      self,
      actual: anndata.AnnData,
      expected: anndata.AnnData,
  ) -> None:
    """Helper function to compare AnnData objects."""
    pd.testing.assert_frame_equal(actual.obs, expected.obs, check_like=True)
    pd.testing.assert_frame_equal(actual.var, expected.var, check_like=True)
    self.assertEqual(actual.uns['interval'], expected.uns['interval'])
    if 'variant' in expected.uns:
      self.assertEqual(actual.uns['variant'], expected.uns['variant'])
    np.testing.assert_array_equal(actual.X, expected.X)
    if 'quantiles' in expected.layers:
      np.testing.assert_array_equal(
          actual.layers['quantiles'], expected.layers['quantiles']
      )
    else:
      self.assertNotIn('quantiles', actual.layers)

  def setUp(self):
    super().setUp()
    self.addTypeEqualityFunc(dna_client.Output, self._assert_output_equal)
    self.addTypeEqualityFunc(
        track_data.TrackData, self._assert_track_data_equal
    )
    self.addTypeEqualityFunc(
        junction_data.JunctionData, self._assert_junction_data_equal
    )
    self.addTypeEqualityFunc(
        dna_client.OutputMetadata, self._assert_output_metadata_equal
    )

  @parameterized.product(
      (
          dict(
              requested_output_types=[*dna_client.OutputType],
              sequence_length=16_384,
              expected=dna_client.Output(
                  atac=track_data.TrackData(
                      values=np.zeros((16_384, 32), dtype=np.float32),
                      metadata=pd.DataFrame({
                          'name': [f'{i}' for i in range(32)],
                          'strand': ['.'] * 32,
                      }),
                  ),
                  rna_seq=track_data.TrackData(
                      values=np.zeros((16_384, 64), dtype=np.float32),
                      metadata=pd.DataFrame({
                          'name': [f'{i}' for i in range(64)],
                          'strand': ['.'] * 64,
                      }),
                  ),
                  splice_junctions=junction_data.JunctionData(
                      junctions=np.array([
                          genome.Interval('chr1', 0, 1000, strand='+'),
                          genome.Interval('chr1', 1500, 2000, strand='-'),
                      ]),
                      values=np.zeros([2, 3], dtype=np.float32),
                      metadata=pd.DataFrame(
                          {'name': [f'{i}' for i in range(3)]}
                      ),
                  ),
              ),
          ),
          dict(
              requested_output_types={dna_client.OutputType.RNA_SEQ},
              sequence_length=16_384,
              expected=dna_client.Output(
                  rna_seq=track_data.TrackData(
                      values=np.zeros((16_384, 64), dtype=np.float32),
                      metadata=pd.DataFrame({
                          'name': [f'{i}' for i in range(64)],
                          'strand': ['.'] * 64,
                      }),
                  ),
              ),
          ),
          dict(
              requested_output_types={dna_client.OutputType.CONTACT_MAPS},
              sequence_length=16_384,
              expected=dna_client.Output(),
          ),
      ),
      bytes_per_chunk=[0, 1024],
      model_version=[None, dna_client.ModelVersion.FOLD_0],
      with_interval=[True, False],
  )
  def test_predict_sequence(
      self,
      requested_output_types: set[dna_client.OutputType] | None,
      sequence_length: int,
      expected: dna_client.Output,
      bytes_per_chunk: int,
      model_version: dna_client.ModelVersion | None,
      with_interval: bool,
  ):
    mock_predictions = _create_example_model_predictions(
        sequence_length=sequence_length
    )

    expected_request = dna_model_service_pb2.PredictSequenceRequest(
        sequence='A' * sequence_length,
        organism=dna_model_pb2.ORGANISM_HOMO_SAPIENS,
        requested_outputs=[o.to_proto() for o in requested_output_types],
        ontology_terms=[
            dna_model_pb2.OntologyTerm(
                ontology_type=dna_model_pb2.ONTOLOGY_TYPE_UBERON, id=4
            )
        ],
        model_version=model_version.name if model_version else None,
    )

    interval = None
    if with_interval:
      interval = genome.Interval('chr1', 0, sequence_length)
      expected_dict = {}
      for field in dataclasses.fields(expected):
        output_type = field.metadata['output_type']
        if output := expected.get(output_type):
          expected_dict[field.name] = dataclasses.replace(
              output, interval=interval
          )
      expected = dna_client.Output(**expected_dict)

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        self.assertEqual(request, expected_request)
        yield from _generate_prediction_protos(
            mock_predictions,
            request.requested_outputs,
            bytes_per_chunk,
            dna_model_service_pb2.PredictSequenceResponse,
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(
        channel=mock_channel, model_version=model_version
    )

    outputs = model.predict_sequence(
        'A' * sequence_length,
        organism=dna_client.Organism.HOMO_SAPIENS,
        ontology_terms=['UBERON:00004'],
        requested_outputs=requested_output_types,
        interval=interval,
    )

    self.assertEqual(outputs, expected)
    mock_stream.assert_called_once()

  @parameterized.product(
      sequence_length=[dna_client.SEQUENCE_LENGTH_16KB],
      num_predictions=[1, 4],
      num_workers=[1, 4],
      bytes_per_chunk=[0, 1024],
      with_interval=[True, False],
  )
  def test_predict_sequences(
      self,
      sequence_length: int,
      num_predictions: int,
      num_workers: int,
      bytes_per_chunk: int,
      with_interval: bool,
  ):
    interval = None
    if with_interval:
      interval = genome.Interval('chr1', 0, sequence_length)

    mock_predictions = _create_example_model_predictions(
        sequence_length=sequence_length, interval=interval
    )

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_prediction_protos(
            mock_predictions,
            request.requested_outputs,
            bytes_per_chunk,
            dna_model_service_pb2.PredictSequenceResponse,
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)

    outputs = model.predict_sequences(
        ['A' * sequence_length] * num_predictions,
        organism=dna_client.Organism.HOMO_SAPIENS,
        ontology_terms=None,
        requested_outputs=[*dna_client.OutputType],
        progress_bar=False,
        max_workers=num_workers,
        intervals=[interval] * num_predictions if with_interval else None,
    )
    for output in outputs:
      self.assertEqual(output, mock_predictions)

    self.assertEqual(mock_stream.call_count, num_predictions)

  def test_predict_sequences_error_raises(self):
    mock_channel, mock_stream = _create_mock_channel_and_stream(
        _FakeRpcError(grpc.StatusCode.INTERNAL)
    )
    model = dna_client.DnaClient(channel=mock_channel)
    with self.assertRaises(grpc.RpcError):
      model.predict_sequences(
          ['A' * dna_client.SEQUENCE_LENGTH_16KB],
          ontology_terms=None,
          requested_outputs=[*dna_client.OutputType],
      )

    mock_stream.assert_called_once()

  def test_invalid_sequence_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError,
        'Invalid sequence, must only contain the characters "ACGTN". Found'
        ' invalid characters: "[wozers,]*"',
    ):
      model.predict_sequence(
          'wowzers',
          ontology_terms=None,
          requested_outputs=[dna_client.OutputType.ATAC],
      )

  @parameterized.product(
      (
          dict(
              requested_output_types={dna_client.OutputType.ATAC},
              interval=genome.Interval('chr1', 0, 16_384),
              expected=dna_client.Output(
                  atac=track_data.TrackData(
                      values=np.zeros((16_384, 32), dtype=np.float32),
                      metadata=pd.DataFrame({
                          'name': [f'{i}' for i in range(32)],
                          'strand': ['.'] * 32,
                      }),
                      interval=genome.Interval('chr1', 0, 16_384),
                  )
              ),
          ),
      ),
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
      model_version=[None, dna_client.ModelVersion.FOLD_0],
  )
  def test_predict_interval(
      self,
      requested_output_types,
      interval,
      organism,
      expected,
      bytes_per_chunk,
      model_version,
  ):
    mock_predictions = _create_example_model_predictions(interval=interval)

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_prediction_protos(
            mock_predictions,
            request.requested_outputs,
            bytes_per_chunk,
            dna_model_service_pb2.PredictIntervalResponse,
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(
        channel=mock_channel, model_version=model_version
    )

    outputs = model.predict_interval(
        interval=interval,
        organism=organism,
        ontology_terms=['UBERON:00004'],
        requested_outputs=requested_output_types,
    )
    self.assertEqual(expected, outputs)

    mock_stream.assert_called_once()

  @parameterized.product(
      intervals=[
          [genome.Interval('chr1', 0, 16_384)],
          [
              genome.Interval('chr1', 0, 16_384),
              genome.Interval('chr1', 16_384, 32_768),
          ],
      ],
      num_workers=[1, 4],
      bytes_per_chunk=[0, 1024],
  )
  def test_predict_intervals(self, intervals, num_workers, bytes_per_chunk):
    examples = {
        (interval.start, interval.end): _create_example_model_predictions(
            interval=interval, value=interval.start
        )
        for interval in intervals
    }

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_prediction_protos(
            examples[(request.interval.start, request.interval.end)],
            request.requested_outputs,
            bytes_per_chunk,
            dna_model_service_pb2.PredictIntervalResponse,
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)

    outputs = model.predict_intervals(
        intervals=intervals,
        ontology_terms=None,
        requested_outputs=[*dna_client.OutputType],
        max_workers=num_workers,
        progress_bar=False,
    )
    for interval, output in zip(intervals, outputs):
      self.assertEqual(output, examples[(interval.start, interval.end)])

    self.assertLen(intervals, mock_stream.call_count)

  def test_predict_intervals_error_raises(self):
    mock_channel, mock_stream = _create_mock_channel_and_stream(
        _FakeRpcError(grpc.StatusCode.INTERNAL)
    )
    model = dna_client.DnaClient(channel=mock_channel)
    with self.assertRaises(grpc.RpcError):
      model.predict_intervals(
          [genome.Interval('chr1', 0, 16_384)],
          ontology_terms=None,
          requested_outputs=[*dna_client.OutputType],
      )

    mock_stream.assert_called_once()

  @parameterized.product(
      (
          dict(
              requested_output_types={dna_client.OutputType.ATAC},
              interval=genome.Interval('chr1', 0, 16_384),
              variant=genome.Variant('chr1', 1000, 'A', 'C'),
              expected=(
                  dna_client.Output(
                      atac=track_data.TrackData(
                          values=np.zeros((16_384, 32), dtype=np.float32),
                          metadata=pd.DataFrame({
                              'name': [f'{i}' for i in range(32)],
                              'strand': ['.'] * 32,
                          }),
                          interval=genome.Interval('chr1', 0, 16_384),
                      )
                  ),
                  dna_client.Output(
                      atac=track_data.TrackData(
                          values=np.ones((16_384, 32), dtype=np.float32),
                          metadata=pd.DataFrame({
                              'name': [f'{i}' for i in range(32)],
                              'strand': ['.'] * 32,
                          }),
                          interval=genome.Interval('chr1', 0, 16_384),
                      )
                  ),
              ),
          ),
      ),
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
      model_version=[None, dna_client.ModelVersion.FOLD_1],
  )
  def test_predict_variant(
      self,
      requested_output_types,
      interval,
      variant,
      organism,
      expected,
      bytes_per_chunk,
      model_version,
  ):
    mock_predictions = dna_client.VariantOutput(
        reference=_create_example_model_predictions(interval=interval, value=0),
        alternate=_create_example_model_predictions(interval=interval, value=1),
    )
    expected = dna_client.VariantOutput(
        reference=expected[0],
        alternate=expected[1],
    )

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_variant_protos(
            mock_predictions, request.requested_outputs, bytes_per_chunk
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(
        channel=mock_channel, model_version=model_version
    )

    outputs = model.predict_variant(
        interval=interval,
        variant=variant,
        organism=organism,
        ontology_terms=['UBERON:00004'],
        requested_outputs=requested_output_types,
    )

    self.assertEqual(expected.reference, outputs.reference)
    self.assertEqual(expected.alternate, outputs.alternate)

    mock_stream.assert_called_once()

  @parameterized.product(
      intervals=[
          genome.Interval('chr1', 0, 16_384),
          [
              genome.Interval('chr1', 0, 16_384),
              genome.Interval('chr1', 16_384, 32_768),
          ],
      ],
      variants=[
          [
              genome.Variant('chr1', 1000, 'A', 'C'),
              genome.Variant('chr1', 2000, 'A', 'C'),
          ],
      ],
      num_workers=[1, 4],
      bytes_per_chunk=[0, 2048],
  )
  def test_predict_variants(
      self, intervals, variants, num_workers, bytes_per_chunk
  ):
    if isinstance(intervals, genome.Interval):
      intervals = [intervals] * len(variants)
    examples = {
        (interval.start, interval.end): _create_example_model_predictions(
            interval=interval, value=interval.start
        )
        for interval in intervals
    }

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        key = (request.interval.start, request.interval.end)
        yield from _generate_variant_protos(
            dna_client.VariantOutput(
                reference=examples[key], alternate=examples[key]
            ),
            request.requested_outputs,
            bytes_per_chunk,
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)

    outputs = model.predict_variants(
        intervals=intervals,
        variants=variants,
        ontology_terms=None,
        requested_outputs=[*dna_client.OutputType],
        max_workers=num_workers,
        progress_bar=False,
    )
    for interval, output in zip(intervals, outputs, strict=True):
      expected = examples[(interval.start, interval.end)]
      self.assertEqual(output.reference, expected)
      self.assertEqual(output.alternate, expected)

    self.assertLen(intervals, mock_stream.call_count)

  def test_predict_variants_error_raises(self):
    mock_channel, mock_stream = _create_mock_channel_and_stream(
        _FakeRpcError(grpc.StatusCode.INTERNAL)
    )
    model = dna_client.DnaClient(channel=mock_channel)
    with self.assertRaises(grpc.RpcError):
      model.predict_variants(
          [genome.Interval('chr1', 0, 16_384)],
          [genome.Variant('chr1', 1000, 'A', 'C')],
          ontology_terms=None,
          requested_outputs=[*dna_client.OutputType],
      )

    mock_stream.assert_called_once()

  def test_predict_variants_mismatch_lengths_raises(self):
    mock_channel, mock_stream = _create_mock_channel_and_stream(None)
    model = dna_client.DnaClient(channel=mock_channel)
    with self.assertRaisesRegex(
        ValueError, 'Intervals and variants must have the same length'
    ):
      model.predict_variants(
          [
              genome.Interval('chr1', 0, 16_384),
              genome.Interval('chr1', 16_384, 32_768),
          ],
          [genome.Variant('chr1', 1000, 'A', 'C')],
          ontology_terms=None,
          requested_outputs=[*dna_client.OutputType],
      )
    mock_stream.assert_not_called()

  def test_output_metadata(self):

    def _mock_generate(requests, metadata):
      del metadata
      del requests
      yield dna_model_service_pb2.MetadataResponse(
          output_metadata=[
              dna_model_pb2.OutputMetadata(
                  output_type=dna_model_pb2.OUTPUT_TYPE_ATAC,
                  tracks=dna_model_pb2.TracksMetadata(
                      metadata=[
                          dna_model_pb2.TrackMetadata(
                              name=f'track_{i}',
                              strand=dna_model_pb2.STRAND_UNSTRANDED,
                          )
                          for i in range(32)
                      ]
                  ),
              ),
              dna_model_pb2.OutputMetadata(
                  output_type=dna_model_pb2.OUTPUT_TYPE_SPLICE_JUNCTIONS,
                  junctions=dna_model_pb2.JunctionsMetadata(
                      metadata=[
                          dna_model_pb2.JunctionMetadata(name=f'junction_{i}')
                          for i in range(16)
                      ]
                  ),
              ),
          ]
      )

    mock_channel, mock_stream = _create_mock_channel_and_stream(
        _mock_generate, bidi_stream=False
    )
    model = dna_client.DnaClient(channel=mock_channel)

    outputs = model.output_metadata(organism=dna_client.Organism.HOMO_SAPIENS)
    expected = dna_client.OutputMetadata(
        atac=track_data.TrackMetadata(
            {'name': [f'track_{i}' for i in range(32)], 'strand': ['.'] * 32}
        ),
        splice_junctions=junction_data.JunctionMetadata(
            {'name': [f'junction_{i}' for i in range(16)]}
        ),
    )
    self.assertEqual(expected, outputs)
    mock_stream.assert_called_once()

  def test_output_type_to_proto(self):
    expected = [
        dna_model_pb2.OUTPUT_TYPE_ATAC,
        dna_model_pb2.OUTPUT_TYPE_CAGE,
        dna_model_pb2.OUTPUT_TYPE_DNASE,
        dna_model_pb2.OUTPUT_TYPE_RNA_SEQ,
        dna_model_pb2.OUTPUT_TYPE_CHIP_HISTONE,
        dna_model_pb2.OUTPUT_TYPE_CHIP_TF,
        dna_model_pb2.OUTPUT_TYPE_SPLICE_SITES,
        dna_model_pb2.OUTPUT_TYPE_SPLICE_SITE_USAGE,
        dna_model_pb2.OUTPUT_TYPE_SPLICE_JUNCTIONS,
        dna_model_pb2.OUTPUT_TYPE_CONTACT_MAPS,
        dna_model_pb2.OUTPUT_TYPE_PROCAP,
    ]

    for output_type, expected_proto in zip(
        dna_client.OutputType, expected, strict=True
    ):
      self.assertEqual(output_type.to_proto(), expected_proto)

  def test_organism_to_proto(self):
    expected = [
        dna_model_pb2.ORGANISM_HOMO_SAPIENS,
        dna_model_pb2.ORGANISM_MUS_MUSCULUS,
    ]

    for organism, expected_proto in zip(
        dna_client.Organism, expected, strict=True
    ):
      self.assertEqual(organism.to_proto(), expected_proto)

  @parameterized.product(
      (
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              scorers=[
                  interval_scorers.GeneMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=2_001,
                      aggregation_type=(
                          interval_scorers.IntervalAggregationType.MEAN
                      ),
                  ),
                  interval_scorers.GeneMaskScorer(
                      requested_output=dna_client.OutputType.DNASE,
                      width=501,
                      aggregation_type=(
                          interval_scorers.IntervalAggregationType.SUM
                      ),
                  ),
              ],
          ),
      ),
      expected=[
          (
              anndata.AnnData(
                  X=np.zeros((5, 10)),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'interval_scorer': interval_scorers.GeneMaskScorer(
                          requested_output=dna_client.OutputType.ATAC,
                          width=2_001,
                          aggregation_type=(
                              interval_scorers.IntervalAggregationType.MEAN
                          ),
                      ),
                  },
              ),
              anndata.AnnData(
                  X=np.ones((5, 10)),
                  obs=pd.DataFrame({
                      'gene_id': [f'{i}' for i in range(5)],
                      'strand': ['-'] * 5,
                      'gene_name': [f'gene_{i}' for i in range(5)],
                      'gene_type': [f'gene_type_{i}' for i in range(5)],
                  }),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'interval_scorer': interval_scorers.GeneMaskScorer(
                          requested_output=dna_client.OutputType.DNASE,
                          width=501,
                          aggregation_type=(
                              interval_scorers.IntervalAggregationType.SUM
                          ),
                      ),
                  },
              ),
          ),
      ],
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
  )
  def test_score_interval(
      self,
      interval,
      scorers,
      expected,
      organism,
      bytes_per_chunk,
  ):

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_interval_scoring_protos(
            scores=expected,
            bytes_per_chunk=bytes_per_chunk,
            response_proto_type=dna_model_service_pb2.ScoreIntervalResponse,
            interval=genome.Interval.from_proto(request.interval),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    outputs = model.score_interval(
        interval=interval,
        organism=organism,
        interval_scorers=scorers,
    )

    for expected_anndata, output_anndata in zip(expected, outputs):
      self._assert_anndata_equal(expected_anndata, output_anndata)

    mock_stream.assert_called_once()

  @parameterized.product(organism=list(dna_client.Organism))
  def test_recommended_interval_scorers(self, organism):
    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_interval_scoring_protos(
            scores=[
                anndata.AnnData(
                    np.zeros((0, 0)),
                    var=pd.DataFrame(columns=['name', 'strand']),
                )
                for _ in range(len(request.interval_scorers))
            ],
            bytes_per_chunk=1024,
            response_proto_type=dna_model_service_pb2.ScoreIntervalResponse,
            interval=genome.Interval.from_proto(request.interval),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    scores = model.score_interval(
        interval=genome.Interval('chr1', 0, 2**20), organism=organism
    )

    self.assertSameElements(
        [score.uns['interval_scorer'] for score in scores],
        interval_scorers.RECOMMENDED_INTERVAL_SCORERS.values(),
    )
    mock_stream.assert_called_once()

  def test_duplicate_interval_scorers_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError, 'Duplicate interval scorers requested'
    ):
      model.score_interval(
          interval=genome.Interval('chr1', 0, 16_384),
          interval_scorers=[
              interval_scorers.GeneMaskScorer(
                  requested_output=dna_client.OutputType.DNASE,
                  width=501,
                  aggregation_type=(
                      interval_scorers.IntervalAggregationType.SUM
                  ),
              )
          ]
          * 2,
      )

  def test_invalid_width_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(ValueError, 'must have widths <= interval'):
      model.score_interval(
          interval=genome.Interval('chr1', 0, 16_384),
          interval_scorers=[
              interval_scorers.GeneMaskScorer(
                  requested_output=dna_client.OutputType.DNASE,
                  width=100_001,
                  aggregation_type=(
                      interval_scorers.IntervalAggregationType.SUM
                  ),
              )
          ],
      )

  def test_exceed_max_interval_scorers_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError, 'Too many interval scorers requested'
    ):
      scorers = []
      for output_type in interval_scorers.SUPPORTED_OUTPUT_TYPES[
          interval_scorers.BaseIntervalScorer.GENE_MASK
      ]:
        for width in interval_scorers.SUPPORTED_WIDTHS[
            interval_scorers.BaseIntervalScorer.GENE_MASK
        ]:
          scorers.append(
              interval_scorers.GeneMaskScorer(
                  requested_output=output_type,
                  width=width,
                  aggregation_type=(
                      interval_scorers.IntervalAggregationType.SUM
                  ),
              )
          )
      model.score_interval(
          interval=genome.Interval('chr1', 0, 16_384), interval_scorers=scorers
      )

  @parameterized.product(
      (
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              scorers=[
                  interval_scorers.GeneMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=2_001,
                      aggregation_type=(
                          interval_scorers.IntervalAggregationType.MEAN
                      ),
                  ),
                  interval_scorers.GeneMaskScorer(
                      requested_output=dna_client.OutputType.DNASE,
                      width=501,
                      aggregation_type=(
                          interval_scorers.IntervalAggregationType.SUM
                      ),
                  ),
              ],
          ),
      ),
      expected=[
          (
              anndata.AnnData(
                  X=np.zeros((5, 10)),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'interval_scorer': interval_scorers.GeneMaskScorer(
                          requested_output=dna_client.OutputType.ATAC,
                          width=2_001,
                          aggregation_type=(
                              interval_scorers.IntervalAggregationType.MEAN
                          ),
                      ),
                  },
              ),
              anndata.AnnData(
                  X=np.ones((5, 10)),
                  obs=pd.DataFrame({
                      'gene_id': [f'{i}' for i in range(5)],
                      'strand': ['-'] * 5,
                      'gene_name': [f'gene_{i}' for i in range(5)],
                      'gene_type': [f'gene_type_{i}' for i in range(5)],
                  }),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'interval_scorer': interval_scorers.GeneMaskScorer(
                          requested_output=dna_client.OutputType.DNASE,
                          width=501,
                          aggregation_type=(
                              interval_scorers.IntervalAggregationType.SUM
                          ),
                      ),
                  },
              ),
          ),
      ],
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
  )
  def test_score_intervals(
      self,
      interval,
      scorers,
      expected,
      organism,
      bytes_per_chunk,
  ):
    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_interval_scoring_protos(
            scores=expected,
            bytes_per_chunk=bytes_per_chunk,
            response_proto_type=dna_model_service_pb2.ScoreIntervalResponse,
            interval=genome.Interval.from_proto(request.interval),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    all_outputs = model.score_intervals(
        intervals=[interval, interval],
        organism=organism,
        interval_scorers=scorers,
    )

    for outputs in all_outputs:
      for expected_anndata, output_anndata in zip(expected, outputs):
        self._assert_anndata_equal(expected_anndata, output_anndata)

    self.assertEqual(mock_stream.call_count, 2)

  @parameterized.product(
      (
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              variant=genome.Variant('chr1', 1000, 'A', 'C'),
              scorers=[
                  variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=10_001,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
                  variant_scorers.GeneMaskLFCScorer(
                      requested_output=dna_client.OutputType.RNA_SEQ,
                  ),
              ],
          ),
      ),
      expected=[
          (
              anndata.AnnData(
                  X=np.zeros((5, 10)),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'variant': genome.Variant('chr1', 1000, 'A', 'C'),
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'variant_scorer': variant_scorers.CenterMaskScorer(
                          requested_output=dna_client.OutputType.ATAC,
                          width=10_001,
                          aggregation_type=(
                              variant_scorers.AggregationType.DIFF_MEAN
                          ),
                      ),
                  },
              ),
              anndata.AnnData(
                  X=np.ones((5, 10)),
                  obs=pd.DataFrame({
                      'gene_id': [f'{i}' for i in range(5)],
                      'strand': ['-'] * 5,
                      'gene_name': [f'gene_{i}' for i in range(5)],
                      'gene_type': [f'gene_type_{i}' for i in range(5)],
                  }),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'variant': genome.Variant('chr1', 1000, 'A', 'C'),
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'variant_scorer': variant_scorers.GeneMaskLFCScorer(
                          requested_output=dna_client.OutputType.RNA_SEQ,
                      ),
                  },
                  layers={'quantiles': np.zeros((5, 10))},
              ),
          ),
          (
              anndata.AnnData(
                  X=np.zeros((5, 10)),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'variant': genome.Variant('chr1', 1000, 'A', 'C'),
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'variant_scorer': variant_scorers.CenterMaskScorer(
                          requested_output=dna_client.OutputType.ATAC,
                          width=10_001,
                          aggregation_type=(
                              variant_scorers.AggregationType.DIFF_MEAN
                          ),
                      ),
                  },
              ),
              anndata.AnnData(
                  X=np.ones((5, 10)),
                  var=pd.DataFrame({
                      'name': [f'{i}' for i in range(10)],
                      'strand': ['.'] * 10,
                  }),
                  uns={
                      'variant': genome.Variant('chr1', 1000, 'A', 'C'),
                      'interval': genome.Interval('chr1', 0, 16_384),
                      'variant_scorer': variant_scorers.GeneMaskLFCScorer(
                          requested_output=dna_client.OutputType.RNA_SEQ,
                      ),
                  },
              ),
          ),
      ],
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
  )
  def test_score_variant(
      self,
      interval,
      variant,
      scorers,
      expected,
      organism,
      bytes_per_chunk,
  ):

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_variant_scoring_protos(
            scores=expected,
            bytes_per_chunk=bytes_per_chunk,
            response_proto_type=dna_model_service_pb2.ScoreVariantResponse,
            variant=genome.Variant.from_proto(request.variant),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    outputs = model.score_variant(
        interval=interval,
        variant=variant,
        organism=organism,
        variant_scorers=scorers,
    )

    for expected_anndata, output_anndata in zip(expected, outputs):
      self._assert_anndata_equal(expected_anndata, output_anndata)

    mock_stream.assert_called_once()

  def test_duplicate_variant_scorers_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError, 'Duplicate variant scorers requested'
    ):
      model.score_variant(
          interval=genome.Interval('chr1', 0, 16_384),
          variant=genome.Variant('chr1', 1000, 'A', 'C'),
          variant_scorers=[
              variant_scorers.GeneMaskLFCScorer(
                  requested_output=dna_client.OutputType.RNA_SEQ,
              ),
              variant_scorers.GeneMaskLFCScorer(
                  requested_output=dna_client.OutputType.RNA_SEQ,
              ),
          ],
      )

  def test_invalid_organism_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesWithLiteralMatch(
        ValueError,
        'Unsupported organism: MUS_MUSCULUS for scorer PolyadenylationScorer().'
        " Supported organisms: ['HOMO_SAPIENS']",
    ):
      model.score_variant(
          interval=genome.Interval('chr1', 0, 16_384),
          variant=genome.Variant('chr1', 1000, 'A', 'C'),
          variant_scorers=[variant_scorers.PolyadenylationScorer()],
          # The PaQTL scorer doesn't currently support mouse.
          organism=dna_client.Organism.MUS_MUSCULUS,
      )

  def test_exceed_max_variant_scorers_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError, 'Too many variant scorers requested'
    ):
      variant_scorers_to_use = []
      for output_type in variant_scorers.SUPPORTED_OUTPUT_TYPES[
          variant_scorers.BaseVariantScorer.CENTER_MASK
      ]:
        for width in variant_scorers.SUPPORTED_WIDTHS[
            variant_scorers.BaseVariantScorer.CENTER_MASK
        ]:
          variant_scorers_to_use.append(
              variant_scorers.CenterMaskScorer(
                  requested_output=output_type,
                  width=width,
                  aggregation_type=variant_scorers.AggregationType.DIFF_MEAN,
              )
          )
      model.score_variant(
          interval=genome.Interval('chr1', 0, 16_384),
          variant=genome.Variant('chr1', 1000, 'A', 'C'),
          variant_scorers=variant_scorers_to_use,
      )

  @parameterized.product(organism=list(dna_client.Organism))
  def test_recommended_variant_scorers(self, organism):
    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_variant_scoring_protos(
            scores=[
                anndata.AnnData(
                    np.zeros((0, 0)),
                    var=pd.DataFrame(columns=['name', 'strand']),
                )
                for _ in range(len(request.variant_scorers))
            ],
            bytes_per_chunk=1024,
            response_proto_type=dna_model_service_pb2.ScoreVariantResponse,
            variant=genome.Variant.from_proto(request.variant),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    scores = model.score_variant(
        interval=genome.Interval('chr1', 0, 16_384),
        variant=genome.Variant('chr1', 1000, 'A', 'C'),
        organism=organism,
    )

    self.assertSameElements(
        [score.uns['variant_scorer'] for score in scores],
        variant_scorers.get_recommended_scorers(organism.to_proto()),
    )
    mock_stream.assert_called_once()

  @parameterized.product(
      (
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              scorers=[
                  variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=10_001,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
                  variant_scorers.GeneMaskLFCScorer(
                      requested_output=dna_client.OutputType.RNA_SEQ,
                  ),
                  variant_scorers.SpliceJunctionScorer(),
              ],
          ),
      ),
      expected=[(
          anndata.AnnData(
              X=np.ones((5, 10)),
              obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
              var=pd.DataFrame({
                  'name': [f'{i}' for i in range(10)],
                  'strand': ['.'] * 10,
              }),
              uns={
                  'interval': genome.Interval('chr1', 0, 16_384),
                  'variant_scorer': variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=10_001,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
              },
          ),
          anndata.AnnData(
              X=np.ones((5, 10)),
              obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
              var=pd.DataFrame({
                  'name': [f'{i}' for i in range(10)],
                  'strand': ['.'] * 10,
              }),
              uns={
                  'interval': genome.Interval('chr1', 0, 16_384),
                  'variant_scorer': variant_scorers.GeneMaskLFCScorer(
                      requested_output=dna_client.OutputType.RNA_SEQ,
                  ),
              },
          ),
          anndata.AnnData(
              X=np.ones((5, 10)),
              obs=pd.DataFrame({
                  'gene_id': [f'{i}' for i in range(5)],
                  'junction_Start': list(range(5)),
                  'junction_End': list(range(1, 6)),
              }),
              var=pd.DataFrame({
                  'name': [f'{i}' for i in range(10)],
                  'strand': ['.'] * 10,
              }),
              uns={
                  'interval': genome.Interval('chr1', 0, 16_384),
                  'variant_scorer': variant_scorers.SpliceJunctionScorer(),
              },
          ),
      )],
      variants=[
          [
              genome.Variant('chr1', 100, 'G', 'T'),
          ],
          [
              genome.Variant('chr1', 100, 'G', 'T'),
              genome.Variant('chr1', 1000, 'A', 'C'),
              genome.Variant('chr1', 1001, 'A', 'C'),
              genome.Variant('chr1', 1002, 'A', 'C'),
          ],
      ],
      organism=[
          dna_client.Organism.HOMO_SAPIENS,
          dna_client.Organism.MUS_MUSCULUS,
      ],
      bytes_per_chunk=[0, 1024],
  )
  def test_score_variants(
      self,
      interval: genome.Interval,
      variants: list[genome.Variant],
      scorers: list[dna_client.VariantScorerTypes],
      expected: list[anndata.AnnData],
      organism: dna_client.Organism,
      bytes_per_chunk: int,
  ):
    position_to_expected_scores = {}
    for variant in variants:
      scores = []
      for score in expected:
        score = score.copy()
        score.uns['variant'] = variant
        score.X *= variant.position
        scores.append(score)
      position_to_expected_scores[variant.position] = scores

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        yield from _generate_variant_scoring_protos(
            scores=position_to_expected_scores[request.variant.position],
            bytes_per_chunk=bytes_per_chunk,
            response_proto_type=dna_model_service_pb2.ScoreVariantResponse,
            variant=genome.Variant.from_proto(request.variant),
        )

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)
    all_scores = model.score_variants(
        intervals=interval,
        variants=variants,
        organism=organism,
        variant_scorers=scorers,
    )

    for scores in all_scores:
      self.assertLen(scores, len(expected))
      expected = position_to_expected_scores[scores[0].uns['variant'].position]
      for expected_anndata, output_anndata in zip(expected, scores):
        self._assert_anndata_equal(output_anndata, expected_anndata)

    self.assertLen(variants, mock_stream.call_count)

  @parameterized.product(
      (
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              ism_interval=genome.Interval('chr1', 10, 11),
              interval_variant=None,
              scorers=[
                  variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=501,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
                  variant_scorers.PolyadenylationScorer(),
              ],
              example_scores=[
                  anndata.AnnData(
                      X=np.zeros((5, 10)),
                      obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
                      var=pd.DataFrame({
                          'name': [f'{i}' for i in range(10)],
                          'strand': ['.'] * 10,
                      }),
                      uns={
                          'interval': genome.Interval('chr1', 0, 16_384),
                          'variant_scorer': variant_scorers.CenterMaskScorer(
                              requested_output=dna_client.OutputType.ATAC,
                              width=501,
                              aggregation_type=(
                                  variant_scorers.AggregationType.DIFF_MEAN
                              ),
                          ),
                      },
                  ),
              ],
              expected_variants=[
                  genome.Variant('chr1', 11, 'A', 'C'),
                  genome.Variant('chr1', 11, 'A', 'G'),
                  genome.Variant('chr1', 11, 'A', 'T'),
              ],
          ),
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              ism_interval=genome.Interval('chr1', 10, 11),
              interval_variant=genome.Variant('chr1', 11, 'A', 'C'),
              scorers=[
                  variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=501,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
              ],
              example_scores=[
                  anndata.AnnData(
                      X=np.zeros((5, 10)),
                      obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
                      var=pd.DataFrame({
                          'name': [f'{i}' for i in range(10)],
                          'strand': ['.'] * 10,
                      }),
                      uns={
                          'interval': genome.Interval('chr1', 0, 16_384),
                          'variant_scorer': variant_scorers.CenterMaskScorer(
                              requested_output=dna_client.OutputType.ATAC,
                              width=10_001,
                              aggregation_type=(
                                  variant_scorers.AggregationType.DIFF_MEAN
                              ),
                          ),
                      },
                  ),
              ],
              expected_variants=[
                  genome.Variant('chr1', 11, 'C', 'A'),
                  genome.Variant('chr1', 11, 'C', 'G'),
                  genome.Variant('chr1', 11, 'C', 'T'),
              ],
          ),
          dict(
              interval=genome.Interval('chr1', 0, 16_384),
              ism_interval=genome.Interval('chr1', 10, 21),
              interval_variant=None,
              scorers=[
                  variant_scorers.CenterMaskScorer(
                      requested_output=dna_client.OutputType.ATAC,
                      width=501,
                      aggregation_type=(
                          variant_scorers.AggregationType.DIFF_MEAN
                      ),
                  ),
                  variant_scorers.PolyadenylationScorer(),
              ],
              example_scores=[
                  anndata.AnnData(
                      X=np.zeros((5, 10)),
                      obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
                      var=pd.DataFrame({
                          'name': [f'{i}' for i in range(10)],
                          'strand': ['.'] * 10,
                      }),
                      uns={
                          'interval': genome.Interval('chr1', 0, 16_384),
                          'variant_scorer': variant_scorers.CenterMaskScorer(
                              requested_output=dna_client.OutputType.ATAC,
                              width=10_001,
                              aggregation_type=(
                                  variant_scorers.AggregationType.DIFF_MEAN
                              ),
                          ),
                      },
                  ),
                  anndata.AnnData(
                      X=np.ones((5, 10)),
                      obs=pd.DataFrame({'gene_id': [f'{i}' for i in range(5)]}),
                      var=pd.DataFrame({
                          'name': [f'{i}' for i in range(10)],
                          'strand': ['.'] * 10,
                      }),
                      uns={
                          'interval': genome.Interval('chr1', 0, 16_384),
                          'variant_scorer': (
                              variant_scorers.PolyadenylationScorer()
                          ),
                      },
                      layers={'quantiles': np.zeros((5, 10))},
                  ),
              ],
              expected_variants=[
                  genome.Variant('chr1', 11, 'A', 'C'),
                  genome.Variant('chr1', 11, 'A', 'G'),
                  genome.Variant('chr1', 11, 'A', 'T'),
                  genome.Variant('chr1', 12, 'A', 'C'),
                  genome.Variant('chr1', 12, 'A', 'G'),
                  genome.Variant('chr1', 12, 'A', 'T'),
                  genome.Variant('chr1', 13, 'A', 'C'),
                  genome.Variant('chr1', 13, 'A', 'G'),
                  genome.Variant('chr1', 13, 'A', 'T'),
                  genome.Variant('chr1', 14, 'A', 'C'),
                  genome.Variant('chr1', 14, 'A', 'G'),
                  genome.Variant('chr1', 14, 'A', 'T'),
                  genome.Variant('chr1', 15, 'A', 'C'),
                  genome.Variant('chr1', 15, 'A', 'G'),
                  genome.Variant('chr1', 15, 'A', 'T'),
                  genome.Variant('chr1', 16, 'A', 'C'),
                  genome.Variant('chr1', 16, 'A', 'G'),
                  genome.Variant('chr1', 16, 'A', 'T'),
                  genome.Variant('chr1', 17, 'A', 'C'),
                  genome.Variant('chr1', 17, 'A', 'G'),
                  genome.Variant('chr1', 17, 'A', 'T'),
                  genome.Variant('chr1', 18, 'A', 'C'),
                  genome.Variant('chr1', 18, 'A', 'G'),
                  genome.Variant('chr1', 18, 'A', 'T'),
                  genome.Variant('chr1', 19, 'A', 'C'),
                  genome.Variant('chr1', 19, 'A', 'G'),
                  genome.Variant('chr1', 19, 'A', 'T'),
                  genome.Variant('chr1', 20, 'A', 'C'),
                  genome.Variant('chr1', 20, 'A', 'G'),
                  genome.Variant('chr1', 20, 'A', 'T'),
                  genome.Variant('chr1', 21, 'A', 'C'),
                  genome.Variant('chr1', 21, 'A', 'G'),
                  genome.Variant('chr1', 21, 'A', 'T'),
              ],
          ),
      ),
      bytes_per_chunk=[0, 1024],
  )
  def test_score_ism_variant(
      self,
      interval,
      ism_interval,
      scorers,
      bytes_per_chunk: int,
      example_scores,
      expected_variants,
      interval_variant: genome.Variant | None,
  ):

    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        ism_interval = genome.Interval.from_proto(request.ism_interval)
        ism_sequence = 'A' * ism_interval.width
        if request.HasField('interval_variant'):
          ref_variant = genome.Variant.from_proto(request.interval_variant)
          self.assertTrue(ref_variant.is_snv)  # Only SNVs supported in test.
          if ref_variant.reference_overlaps(ism_interval):
            pos = ref_variant.start - ism_interval.start
            ism_sequence = (
                ism_sequence[:pos]
                + ref_variant.alternate_bases
                + ism_sequence[pos + 1 :]
            )

        for variant in ism.ism_variants(ism_interval, ism_sequence):
          yield from _generate_variant_scoring_protos(
              scores=example_scores,
              bytes_per_chunk=bytes_per_chunk,
              response_proto_type=dna_model_service_pb2.ScoreIsmVariantResponse,
              variant=variant,
          )

    mock_channel, _ = _create_mock_channel_and_stream(_mock_generate)

    model = dna_client.DnaClient(channel=mock_channel)
    ism_scores = model.score_ism_variants(
        interval=interval,
        ism_interval=ism_interval,
        variant_scorers=scorers,
        interval_variant=interval_variant,
    )
    self.assertLen(ism_scores, len(expected_variants))
    ism_scores = sorted(ism_scores, key=lambda x: str(x[0].uns['variant']))
    for expected_variant, outputs in zip(
        expected_variants, ism_scores, strict=True
    ):
      for output, example_output in zip(outputs, example_scores, strict=True):
        expected = example_output.copy()
        expected.uns['variant'] = expected_variant
        self._assert_anndata_equal(expected, output)

  def test_score_negative_strand_ism_interval_raises_error(self):
    model = dna_client.DnaClient(channel=mock.create_autospec(grpc.Channel))
    with self.assertRaisesRegex(
        ValueError, 'ISM interval must be on the positive strand.'
    ):
      model.score_ism_variants(
          interval=genome.Interval('chr1', 0, 16_384),
          ism_interval=genome.Interval(
              'chr1', 10, 11, strand=genome.STRAND_NEGATIVE
          ),
          variant_scorers=[],
      )

  def test_create(self):
    with (
        mock.patch.object(grpc, 'secure_channel') as mock_secure_channel,
        mock.patch.object(grpc, 'channel_ready_future'),
    ):
      _ = dna_client.create('foo')
      mock_secure_channel.assert_called_once_with(
          'dns:///gdmscience.googleapis.com:443',
          mock.ANY,
          options=mock.ANY,
      )

  def test_missing_chunk_raises_error(self):
    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        for response in _generate_prediction_protos(
            _create_example_model_predictions(
                sequence_length=len(request.sequence)
            ),
            request.requested_outputs,
            bytes_per_chunk=128,
            response_proto_type=dna_model_service_pb2.PredictSequenceResponse,
        ):
          if response.WhichOneof('payload') != 'tensor_chunk':
            yield response

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)

    with self.assertRaisesRegex(
        ValueError, 'Expected tensor_chunk, got "output" payload'
    ):
      _ = model.predict_sequence(
          sequence='A' * 16_384,
          requested_outputs=[
              dna_client.OutputType.ATAC,
              dna_client.OutputType.RNA_SEQ,
          ],
          ontology_terms=None,
      )

    mock_stream.assert_called_once()

  def test_missing_end_of_stream_chunk_raises_error(self):
    def _mock_generate(requests, metadata):
      del metadata
      for request in requests:
        for response in _generate_prediction_protos(
            _create_example_model_predictions(
                sequence_length=len(request.sequence)
            ),
            request.requested_outputs,
            bytes_per_chunk=128,
            response_proto_type=dna_model_service_pb2.PredictSequenceResponse,
        ):
          if response.WhichOneof('payload') != 'tensor_chunk':
            yield response

    mock_channel, mock_stream = _create_mock_channel_and_stream(_mock_generate)
    model = dna_client.DnaClient(channel=mock_channel)

    with self.assertRaisesRegex(
        ValueError, 'Expected tensor_chunk, got end of stream'
    ):
      _ = model.predict_sequence(
          sequence='A' * 16_384,
          requested_outputs=[dna_client.OutputType.ATAC],
          ontology_terms=None,
      )

    mock_stream.assert_called_once()


class RpcRetryTest(parameterized.TestCase):

  def test_retry_rpc(self):
    mock_channel, _ = _create_mock_channel_and_stream([
        _FakeRpcError(grpc.StatusCode.UNAVAILABLE),
        'bar',
    ])

    @functools.partial(dna_client.retry_rpc, initial_backoff=0.01)
    def foo(channel):
      return channel.stream_stream('method')(iter([]))

    self.assertEqual(foo(mock_channel), 'bar')

  @parameterized.named_parameters(
      {
          'testcase_name': 'normal',
          'initial_backoff': 1.25,
          'backoff_multiplier': 1.5,
          'max_attempts': 5,
          'expected_backoff': [1.25, 1.875, 2.8125, 4.21875],
      },
      {
          'testcase_name': 'no_backoff_multiplier',
          'initial_backoff': 1,
          'backoff_multiplier': 1.0,
          'max_attempts': 2,
          'expected_backoff': [1.0],
      },
      {
          'testcase_name': 'no_retries',
          'initial_backoff': 1,
          'backoff_multiplier': 1.0,
          'max_attempts': 1,
          'expected_backoff': [],
      },
  )
  @mock.patch.object(time, 'sleep', autospec=True)
  def test_retry_fails_after_max_attempts(
      self,
      mock_sleep: mock.Mock,
      initial_backoff: float,
      backoff_multiplier: float,
      max_attempts: int,
      expected_backoff: Sequence[float],
  ):
    mock_channel, _ = _create_mock_channel_and_stream(
        _FakeRpcError(grpc.StatusCode.UNAVAILABLE)
    )

    @functools.partial(
        dna_client.retry_rpc,
        max_attempts=max_attempts,
        initial_backoff=initial_backoff,
        backoff_multiplier=backoff_multiplier,
        jitter=0.0,
    )
    def function(channel: grpc.Channel):
      return dna_model_service_pb2_grpc.DnaModelServiceStub(
          channel
      ).PredictSequence(iter([dna_model_service_pb2.PredictSequenceRequest()]))

    with self.assertRaises(grpc.RpcError):
      _ = function(mock_channel)

    mock_sleep.assert_has_calls(
        [mock.call(backoff) for backoff in expected_backoff]
    )

  def test_invalid_backoff_multiplier_raises_error(self):
    with self.assertRaisesRegex(
        ValueError, 'Backoff multiplier must be >= 1.0.'
    ):

      @functools.partial(dna_client.retry_rpc, backoff_multiplier=0.5)
      def foo():
        return 'bar'

      _ = foo()


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import genome
from alphagenome.interpretation import ism
import numpy as np


class IsmTest(parameterized.TestCase):

  def test_ism_variants(self):
    sequence = 'ACG'
    interval = genome.Interval('chr1', 0, len(sequence))
    variants = ism.ism_variants(interval, sequence=sequence)
    expected = [
        genome.Variant('chr1', 1, 'A', 'C'),
        genome.Variant('chr1', 1, 'A', 'G'),
        genome.Variant('chr1', 1, 'A', 'T'),
        genome.Variant('chr1', 2, 'C', 'A'),
        genome.Variant('chr1', 2, 'C', 'G'),
        genome.Variant('chr1', 2, 'C', 'T'),
        genome.Variant('chr1', 3, 'G', 'A'),
        genome.Variant('chr1', 3, 'G', 'C'),
        genome.Variant('chr1', 3, 'G', 'T'),
    ]
    self.assertEqual(variants, expected)

  def test_ism_variants_length_mismatch(self):
    sequence = 'ACG'
    interval = genome.Interval('chr1', 0, 2)
    with self.assertRaises(ValueError):
      ism.ism_variants(interval, sequence=sequence)

  @parameterized.parameters([True, False])
  def test_ism_variants_with_interval_with_n(self, skip_n):
    sequence = 'NCG'
    interval = genome.Interval('chr1', 0, len(sequence))
    variants = ism.ism_variants(interval, sequence=sequence, skip_n=skip_n)
    expected_n = [
        genome.Variant('chr1', 1, 'N', 'A'),
        genome.Variant('chr1', 1, 'N', 'C'),
        genome.Variant('chr1', 1, 'N', 'G'),
        genome.Variant('chr1', 1, 'N', 'T'),
    ]
    expected = [
        genome.Variant('chr1', 2, 'C', 'A'),
        genome.Variant('chr1', 2, 'C', 'G'),
        genome.Variant('chr1', 2, 'C', 'T'),
        genome.Variant('chr1', 3, 'G', 'A'),
        genome.Variant('chr1', 3, 'G', 'C'),
        genome.Variant('chr1', 3, 'G', 'T'),
    ]
    self.assertEqual(variants, expected if skip_n else expected_n + expected)

  @parameterized.parameters(dict(multiply_by_sequence=[True, False]))
  def test_ism_scores(self, multiply_by_sequence):
    # 'ACG'
    variants = [
        genome.Variant('chr1', 1, 'A', 'C'),
        genome.Variant('chr1', 1, 'A', 'G'),
        genome.Variant('chr1', 1, 'A', 'T'),
        genome.Variant('chr1', 2, 'C', 'A'),
        genome.Variant('chr1', 2, 'C', 'G'),
        genome.Variant('chr1', 2, 'C', 'T'),
        genome.Variant('chr1', 3, 'G', 'A'),
        genome.Variant('chr1', 3, 'G', 'C'),
        genome.Variant('chr1', 3, 'G', 'T'),
    ]
    ism_matrix = ism.ism_matrix(
        variant_scores=[1] * len(variants),
        variants=variants,
        multiply_by_sequence=multiply_by_sequence,
    )
    if multiply_by_sequence:
      expect = np.array(
          [
              [-1, 0, 0, 0],
              [0, -1, 0, 0],
              [0, 0, -1, 0],
          ],
          dtype=np.float32,
      )
      np.testing.assert_array_equal(ism_matrix, expect)
    else:
      expect = (
          np.array(
              [
                  [0, 1, 1, 1],
                  [1, 0, 1, 1],
                  [1, 1, 0, 1],
              ],
              dtype=np.float32,
          )
          - 1
      )
      np.testing.assert_array_equal(ism_matrix, expect)

  def test_ism_scores_subset(self):
    variants = [
        genome.Variant('chr1', 1, 'A', 'C'),
        genome.Variant('chr1', 1, 'A', 'G'),
        genome.Variant('chr1', 1, 'A', 'T'),
        genome.Variant('chr1', 2, 'C', 'A'),
        genome.Variant('chr1', 2, 'C', 'G'),
        genome.Variant('chr1', 2, 'C', 'T'),
        genome.Variant('chr1', 3, 'G', 'A'),
        genome.Variant('chr1', 3, 'G', 'C'),
        genome.Variant('chr1', 3, 'G', 'T'),
    ]
    ism_matrix = ism.ism_matrix(
        variant_scores=[1] * len(variants),
        variants=variants,
        interval=genome.Interval('chr1', 1, 3),
        multiply_by_sequence=False,
    )

    expect = (
        np.array(
            [
                [1, 0, 1, 1],
                [1, 1, 0, 1],
            ],
            dtype=np.float32,
        )
        - 1
    )
    np.testing.assert_array_equal(ism_matrix, expect)

  def test_ism_scores_missing_variants(self):
    variants = [
        genome.Variant('chr1', 1, 'A', 'C'),
        genome.Variant('chr1', 1, 'A', 'G'),
        genome.Variant('chr1', 1, 'A', 'T'),
        genome.Variant('chr1', 2, 'C', 'A'),
        genome.Variant('chr1', 2, 'C', 'G'),
        genome.Variant('chr1', 2, 'C', 'T'),
        genome.Variant('chr1', 3, 'G', 'A'),
        genome.Variant('chr1', 3, 'G', 'C'),
        genome.Variant('chr1', 3, 'G', 'T'),
    ]
    with self.assertRaises(ValueError):
      # Out of bounds.
      ism.ism_matrix(
          variant_scores=[1] * len(variants),
          variants=variants,
          interval=genome.Interval('chr1', 1, 4),
          multiply_by_sequence=False,
      )

    output = ism.ism_matrix(
        variant_scores=[1] * len(variants),
        variants=variants,
        interval=genome.Interval('chr1', 1, 4),
        multiply_by_sequence=False,
        require_fully_filled=False,
    )
    np.testing.assert_array_equal(output[-1], 0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Iterable, Sequence
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import genome
from alphagenome.data import ontology
from alphagenome.models import dna_model
from alphagenome.models import dna_output
from alphagenome.models import interval_scorers as interval_scorers_lib
from alphagenome.models import variant_scorers as variant_scorers_lib
import anndata


class _MockDnaModel(dna_model.DnaModel):
  """Mock DNA model for testing."""

  def __init__(
      self,
      *,
      predict_sequence: Callable[..., dna_output.Output] | None = None,
      predict_interval: Callable[..., dna_output.Output] | None = None,
      predict_variant: Callable[..., dna_output.VariantOutput] | None = None,
      score_interval: Callable[..., list[anndata.AnnData]] | None = None,
      score_variant: Callable[..., list[anndata.AnnData]] | None = None,
      score_ism_variants: (
          Callable[..., list[list[anndata.AnnData]]] | None
      ) = None,
      output_metadata: Callable[..., dna_output.OutputMetadata] | None = None,
  ):
    self._predict_sequence_callable = predict_sequence
    self._predict_interval_callable = predict_interval
    self._predict_variant_callable = predict_variant
    self._score_interval_callable = score_interval
    self._score_variant_callable = score_variant
    self._output_metadata_callable = output_metadata
    self._score_ism_variants_callable = score_ism_variants

  def predict_sequence(
      self,
      sequence: str,
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[dna_output.OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
      interval: genome.Interval | None = None,
  ) -> dna_output.Output:
    if self._predict_sequence_callable is None:
      raise NotImplementedError('predict_sequence is not implemented.')
    return self._predict_sequence_callable(
        sequence,
        organism=organism,
        requested_outputs=requested_outputs,
        ontology_terms=ontology_terms,
        interval=interval,
    )

  def predict_interval(
      self,
      interval: genome.Interval,
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[dna_output.OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
  ) -> dna_output.Output:
    if self._predict_interval_callable is None:
      raise NotImplementedError('predict_interval is not implemented.')
    return self._predict_interval_callable(
        interval,
        organism=organism,
        requested_outputs=requested_outputs,
        ontology_terms=ontology_terms,
    )

  def predict_variant(
      self,
      interval: genome.Interval,
      variant: genome.Variant,
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
      requested_outputs: Iterable[dna_output.OutputType],
      ontology_terms: Iterable[ontology.OntologyTerm | str] | None,
  ) -> dna_output.VariantOutput:
    if self._predict_variant_callable is None:
      raise NotImplementedError('predict_variant is not implemented.')
    return self._predict_variant_callable(
        interval,
        variant,
        organism=organism,
        requested_outputs=requested_outputs,
        ontology_terms=ontology_terms,
    )

  def score_variant(
      self,
      interval: genome.Interval,
      variant: genome.Variant,
      variant_scorers: Sequence[variant_scorers_lib.VariantScorerTypes] = (),
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
  ) -> list[anndata.AnnData]:
    if self._score_variant_callable is None:
      raise NotImplementedError('score_variant is not implemented.')
    return self._score_variant_callable(
        interval,
        variant,
        variant_scorers=variant_scorers,
        organism=organism,
    )

  def score_interval(
      self,
      interval: genome.Interval,
      interval_scorers: Sequence[interval_scorers_lib.IntervalScorerTypes] = (),
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
  ) -> list[anndata.AnnData]:
    if self._score_interval_callable is None:
      raise NotImplementedError('score_interval is not implemented.')
    return self._score_interval_callable(
        interval,
        interval_scorers=interval_scorers,
        organism=organism,
    )

  def output_metadata(
      self, organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS
  ) -> dna_output.OutputMetadata:
    if self._output_metadata_callable is None:
      raise NotImplementedError('output_metadata is not implemented.')
    return self._output_metadata_callable(organism)

  def score_ism_variants(
      self,
      interval: genome.Interval,
      ism_interval: genome.Interval,
      variant_scorers: Sequence[variant_scorers_lib.VariantScorerTypes] = (),
      *,
      organism: dna_model.Organism = dna_model.Organism.HOMO_SAPIENS,
  ) -> list[list[anndata.AnnData]]:
    if self._score_ism_variants_callable is None:
      raise NotImplementedError('score_ism_variants is not implemented.')
    return self._score_ism_variants_callable(
        interval,
        ism_interval,
        variant_scorers=variant_scorers,
        organism=organism,
    )


class DnaModelTest(parameterized.TestCase):

  def test_predict_sequences(self):
    mock_predict_sequence = mock.Mock(return_value=dna_output.Output())
    model = _MockDnaModel(predict_sequence=mock_predict_sequence)

    expected_sequences = [letter * 10 for letter in 'ACGT']

    model.predict_sequences(
        sequences=expected_sequences,
        requested_outputs=[dna_output.OutputType.RNA_SEQ],
        ontology_terms=None,
        progress_bar=False,
    )
    for sequence in expected_sequences:
      mock_predict_sequence.assert_any_call(
          sequence,
          organism=dna_model.Organism.HOMO_SAPIENS,
          requested_outputs=[dna_output.OutputType.RNA_SEQ],
          ontology_terms=None,
          interval=None,
      )

  def test_predict_intervals(self):
    mock_predict_interval = mock.Mock(return_value=dna_output.Output())
    model = _MockDnaModel(predict_interval=mock_predict_interval)

    expected_intervals = [
        genome.Interval(chromosome, 0, 1000)
        for chromosome in ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6']
    ]

    model.predict_intervals(
        intervals=expected_intervals,
        requested_outputs=[dna_output.OutputType.ATAC],
        ontology_terms=None,
        organism=dna_model.Organism.MUS_MUSCULUS,
        progress_bar=False,
    )
    for interval in expected_intervals:
      mock_predict_interval.assert_any_call(
          interval,
          organism=dna_model.Organism.MUS_MUSCULUS,
          requested_outputs=[dna_output.OutputType.ATAC],
          ontology_terms=None,
      )

  @parameterized.product(single_interval=[True, False])
  def test_predict_variants(self, single_interval: bool):
    mock_predict_variant = mock.Mock(
        return_value=dna_output.VariantOutput(
            reference=dna_output.Output(), alternate=dna_output.Output()
        )
    )
    model = _MockDnaModel(predict_variant=mock_predict_variant)

    expected_variants = [
        genome.Variant('chr1', i * 2048, 'A', 'C') for i in range(1, 10)
    ]

    if single_interval:
      expected_intervals = [
          expected_variants[0].reference_interval.resize(2048)
      ] * len(expected_variants)
    else:
      expected_intervals = [
          variant.reference_interval.resize(2048)
          for variant in expected_variants
      ]

    model.predict_variants(
        intervals=expected_intervals[0]
        if single_interval
        else expected_intervals,
        variants=expected_variants,
        requested_outputs=[dna_output.OutputType.CAGE],
        ontology_terms=[ontology.from_curie('UBERON:00001')],
        progress_bar=False,
    )
    for variant, interval in zip(expected_variants, expected_intervals):
      mock_predict_variant.assert_any_call(
          interval,
          variant,
          organism=dna_model.Organism.HOMO_SAPIENS,
          requested_outputs=[dna_output.OutputType.CAGE],
          ontology_terms=[ontology.from_curie('UBERON:00001')],
      )

  def test_score_intervals(self):
    mock_score_interval = mock.Mock(return_value=[anndata.AnnData()])
    model = _MockDnaModel(score_interval=mock_score_interval)

    expected_intervals = [
        genome.Interval(chromosome, 0, 1000)
        for chromosome in ['chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6']
    ]

    model.score_intervals(
        intervals=expected_intervals,
        interval_scorers=list(
            interval_scorers_lib.RECOMMENDED_INTERVAL_SCORERS.values()
        ),
        progress_bar=False,
    )
    for interval in expected_intervals:
      mock_score_interval.assert_any_call(
          interval,
          interval_scorers=list(
              interval_scorers_lib.RECOMMENDED_INTERVAL_SCORERS.values()
          ),
          organism=dna_model.Organism.HOMO_SAPIENS,
      )

  @parameterized.product(single_interval=[True, False])
  def test_score_variants(self, single_interval: bool):
    mock_score_variant = mock.Mock(return_value=[anndata.AnnData()])
    model = _MockDnaModel(score_variant=mock_score_variant)

    expected_variants = [
        genome.Variant('chr1', i * 2048, 'A', 'C') for i in range(1, 10)
    ]

    if single_interval:
      expected_intervals = [
          expected_variants[0].reference_interval.resize(2048)
      ] * len(expected_variants)
    else:
      expected_intervals = [
          variant.reference_interval.resize(2048)
          for variant in expected_variants
      ]

    model.score_variants(
        intervals=expected_intervals[0]
        if single_interval
        else expected_intervals,
        variants=expected_variants,
        variant_scorers=list(
            variant_scorers_lib.RECOMMENDED_VARIANT_SCORERS.values()
        ),
        progress_bar=False,
    )
    for variant, interval in zip(expected_variants, expected_intervals):
      mock_score_variant.assert_any_call(
          interval,
          variant,
          variant_scorers=list(
              variant_scorers_lib.RECOMMENDED_VARIANT_SCORERS.values()
          ),
          organism=dna_model.Organism.HOMO_SAPIENS,
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome import tensor_utils
from alphagenome.protos import tensor_pb2
import ml_dtypes
import numpy as np
import zstandard


class TensorUtilsTest(parameterized.TestCase):

  @parameterized.product(
      (
          dict(
              array=np.array([[1, 2], [3, 4]], dtype=ml_dtypes.bfloat16),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_BFLOAT16,
          ),
          dict(
              array=np.array([[1, 2], [3, 4]], dtype=np.float16),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_FLOAT16,
          ),
          dict(
              array=np.array([[1, 2, 3, 4]], dtype=np.float32),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_FLOAT32,
          ),
          dict(
              array=np.array([[1], [2], [3], [4]], dtype=np.float64),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_FLOAT64,
          ),
          dict(
              array=np.array([[1, 2], [3, 4]], dtype=np.int8),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_INT8,
          ),
          dict(
              array=np.array([1, 2, 3, 4], dtype=np.int32),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_INT32,
          ),
          dict(
              array=np.array([1, 2, 3, 4], dtype=np.int64),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_INT64,
          ),
          dict(
              array=np.array([1, 2, 3, 4], dtype=np.uint8),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_UINT8,
          ),
          dict(
              array=np.array([1, 2, 3, 4], dtype=np.uint32),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_UINT32,
          ),
          dict(
              array=np.array([1, 2, 3, 4], dtype=np.uint64),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_UINT64,
          ),
          dict(
              array=np.array([True, False, True, False], dtype=bool),
              expected_dtype=tensor_pb2.DataType.DATA_TYPE_BOOL,
          ),
      ),
      compression_type=[
          tensor_pb2.CompressionType.COMPRESSION_TYPE_NONE,
          tensor_pb2.CompressionType.COMPRESSION_TYPE_ZSTD,
      ],
  )
  def test_pack_tensor(self, array, expected_dtype, compression_type):
    packed, chunks = tensor_utils.pack_tensor(
        array, compression_type=compression_type
    )

    self.assertEmpty(chunks)
    self.assertSequenceEqual(packed.shape, array.shape)
    self.assertEqual(packed.data_type, expected_dtype)

    expected = (
        zstandard.compress(array.tobytes())
        if compression_type == tensor_pb2.CompressionType.COMPRESSION_TYPE_ZSTD
        else array.tobytes()
    )
    self.assertEqual(expected, packed.array.data)

  @parameterized.product(
      dtype=(
          ml_dtypes.bfloat16,
          np.float16,
          np.float32,
          np.float64,
          np.int8,
          np.int32,
          np.int64,
          np.uint8,
          np.uint32,
          np.uint64,
          bool,
      ),
      bytes_per_chunk=[10, 13, 200],
      compression_type=[
          tensor_pb2.CompressionType.COMPRESSION_TYPE_NONE,
          tensor_pb2.CompressionType.COMPRESSION_TYPE_ZSTD,
      ],
  )
  def test_pack_tensor_chunks(self, dtype, bytes_per_chunk, compression_type):
    if dtype == bool:
      data = np.ones(1000, dtype=bool)
    else:
      data = np.arange(1000, dtype=dtype)
    data = data.reshape((10, -1))
    packed, chunks = tensor_utils.pack_tensor(
        data, bytes_per_chunk=bytes_per_chunk, compression_type=compression_type
    )
    self.assertEmpty(packed.array.data)
    self.assertLen(chunks, packed.chunk_count)

    items_per_chunk = bytes_per_chunk // data.itemsize
    self.assertEqual(packed.chunk_count, math.ceil(data.size / items_per_chunk))
    first_chunk = data.flatten()[:items_per_chunk].tobytes()
    expected = (
        zstandard.compress(first_chunk)
        if compression_type == tensor_pb2.CompressionType.COMPRESSION_TYPE_ZSTD
        else first_chunk
    )
    self.assertEqual(expected, chunks[0].data)

  @parameterized.product(
      array=(
          np.array([[1, 2], [3, 4]], dtype=ml_dtypes.bfloat16),
          np.array([[1, 2], [3, 4]], dtype=np.float16),
          np.array([[1, 2, 3, 4]], dtype=np.float32),
          np.array([[1], [2], [3], [4]], dtype=np.float64),
          np.array([[1, 2], [3, 4]], dtype=np.int8),
          np.array([1, 2, 3, 4], dtype=np.int32),
          np.array([1, 2, 3, 4], dtype=np.int64),
          np.array([1, 2, 3, 4], dtype=np.uint8),
          np.array([1, 2, 3, 4], dtype=np.uint32),
          np.array([1, 2, 3, 4], dtype=np.uint64),
          np.array([True, False, True, False], dtype=bool),
          np.arange(100)[::2],
      ),
      bytes_per_chunk=[8, 12, 15],
      compression_type=[
          tensor_pb2.CompressionType.COMPRESSION_TYPE_NONE,
          tensor_pb2.CompressionType.COMPRESSION_TYPE_ZSTD,
      ],
  )
  def test_unpack_proto(self, array, bytes_per_chunk, compression_type):
    tensor, chunks = tensor_utils.pack_tensor(
        array,
        bytes_per_chunk=bytes_per_chunk,
        compression_type=compression_type,
    )
    self.assertLen(chunks, tensor.chunk_count)
    round_trip = tensor_utils.unpack_proto(tensor, chunks)
    np.testing.assert_array_equal(array, round_trip)

  def test_pack_proto_invalid_chunk_size_raises(self):
    with self.assertRaisesRegex(
        ValueError, 'bytes_per_chunk=1 must be >= value.itemsize=8'
    ):
      tensor_utils.pack_tensor(
          np.array([[1, 2], [3, 4]], dtype=np.int64), bytes_per_chunk=1
      )

  def test_missing_chunk_raises(self):
    tensor, chunks = tensor_utils.pack_tensor(
        np.arange(128, dtype=np.float32), bytes_per_chunk=32
    )
    chunks = chunks[::2]
    with self.assertRaisesWithLiteralMatch(
        ValueError, 'cannot reshape array of size 64 into shape (128,)'
    ):
      tensor_utils.unpack_proto(tensor, chunks)

  @parameterized.named_parameters(
      (
          'float16',
          np.array([1, 2, 3], dtype=np.float16),
          np.array([1, 2, 3], dtype=np.float32),
      ),
      (
          'bfloat16',
          np.array([1, 2, 3], dtype=ml_dtypes.bfloat16),
          np.array([1, 2, 3], dtype=np.float32),
      ),
      (
          'float32',
          np.array([1, 2, 3], dtype=np.float32),
          np.array([1, 2, 3], dtype=np.float32),
      ),
      (
          'float64',
          np.array([1, 2, 3], dtype=np.float64),
          np.array([1, 2, 3], dtype=np.float64),
      ),
      (
          'int',
          np.array([1, 2, 3], dtype=np.int32),
          np.array([1, 2, 3], dtype=np.int32),
      ),
  )
  def test_upcast_floating(self, array, expected):
    np.testing.assert_array_equal(tensor_utils.upcast_floating(array), expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import fold_intervals
from alphagenome.models import dna_client
import pandas as pd


_dummy_intervals = pd.DataFrame({
    'chromosome': ['chr' + str(i) for i in range(8)],
    'start': [i for i in range(0, 8000, 1000)],
    'end': [i + 500 for i in range(0, 8000, 1000)],
    'fold': ['fold' + str(i) for i in range(8)],
})


class FoldIntervalsTest(parameterized.TestCase):
  """Tests for the fold intervals."""

  @parameterized.named_parameters(
      dict(
          testcase_name='train_fold_0',
          model_version=dna_client.ModelVersion.FOLD_0,
          subset=fold_intervals.Subset.TRAIN,
          expected_intervals=_dummy_intervals[
              # Folds 0 and 1 are held out
              _dummy_intervals.fold.isin(
                  ['fold' + str(i) for i in [2, 3, 4, 5, 6, 7]]
              )
          ],
      ),
      dict(
          testcase_name='train_fold_1',
          model_version=dna_client.ModelVersion.FOLD_1,
          subset=fold_intervals.Subset.TRAIN,
          expected_intervals=_dummy_intervals[
              # Folds 3 and 4 are held out (aligns with Borzoi fold 1)
              _dummy_intervals.fold.isin(
                  ['fold' + str(i) for i in [0, 1, 2, 5, 6, 7]]
              )
          ],
      ),
      dict(
          testcase_name='train_all_folds',
          model_version=dna_client.ModelVersion.ALL_FOLDS,
          subset=fold_intervals.Subset.TRAIN,
          expected_intervals=_dummy_intervals,
      ),
      dict(
          testcase_name='valid_fold_0',
          model_version=dna_client.ModelVersion.FOLD_0,
          subset=fold_intervals.Subset.VALID,
          expected_intervals=_dummy_intervals[
              # Model fold 0 uses data fold 0 for validation.
              _dummy_intervals.fold
              == 'fold0'
          ],
      ),
      dict(
          testcase_name='valid_fold_3',
          model_version=dna_client.ModelVersion.FOLD_3,
          subset=fold_intervals.Subset.VALID,
          expected_intervals=_dummy_intervals[
              # Model fold 3 uses data fold 6 for validation.
              _dummy_intervals.fold
              == 'fold6'
          ],
      ),
      dict(
          testcase_name='valid_all_folds',
          model_version=dna_client.ModelVersion.ALL_FOLDS,
          subset=fold_intervals.Subset.VALID,
          expected_intervals=_dummy_intervals[
              # All folds model uses data fold 0 for validation.
              _dummy_intervals.fold
              == 'fold0'
          ],
      ),
      dict(
          testcase_name='test_fold_1',
          model_version=dna_client.ModelVersion.FOLD_1,
          subset=fold_intervals.Subset.TEST,
          expected_intervals=_dummy_intervals[
              # Model fold 1 uses data fold 4 for testing.
              _dummy_intervals.fold
              == 'fold4'
          ],
      ),
      dict(
          testcase_name='test_fold_2',
          model_version=dna_client.ModelVersion.FOLD_2,
          subset=fold_intervals.Subset.TEST,
          expected_intervals=_dummy_intervals[
              # Model fold 4 uses data fold 5 for testing.
              _dummy_intervals.fold
              == 'fold5'
          ],
      ),
      dict(
          testcase_name='test_all_folds',
          model_version=dna_client.ModelVersion.ALL_FOLDS,
          subset=fold_intervals.Subset.TEST,
          expected_intervals=_dummy_intervals[
              # All folds model uses data fold 0 for testing.
              _dummy_intervals.fold
              == 'fold1'
          ],
      ),
  )
  def test_get_fold_intervals(
      self,
      model_version: dna_client.ModelVersion,
      subset: fold_intervals.Subset,
      expected_intervals: pd.DataFrame,
  ):
    path = self.create_tempfile(
        content=_dummy_intervals.to_csv(sep='\t', header=False)
    ).full_path
    intervals = fold_intervals.get_fold_intervals(
        model_version=model_version,
        organism=dna_client.Organism.HOMO_SAPIENS,
        subset=subset,
        example_regions_path=path,
    )
    pd.testing.assert_frame_equal(
        intervals,
        expected_intervals,
    )


if __name__ == '__main__':
  absltest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: alphagenome/protos/dna_model.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'alphagenome/protos/dna_model.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from alphagenome.protos import tensor_pb2 as alphagenome_dot_protos_dot_tensor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"alphagenome/protos/dna_model.proto\x12(google.gdm.gdmscience.alphagenome.v1main\x1a\x1f\x61lphagenome/protos/tensor.proto\"|\n\x08Interval\x12\x12\n\nchromosome\x18\x01 \x01(\t\x12\r\n\x05start\x18\x02 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x03 \x01(\x03\x12@\n\x06strand\x18\x04 \x01(\x0e\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Strand\"a\n\x07Variant\x12\x12\n\nchromosome\x18\x01 \x01(\t\x12\x10\n\x08position\x18\x02 \x01(\x03\x12\x17\n\x0freference_bases\x18\x03 \x01(\t\x12\x17\n\x0f\x61lternate_bases\x18\x04 \x01(\t\"i\n\x0cOntologyTerm\x12M\n\rontology_type\x18\x01 \x01(\x0e\x32\x36.google.gdm.gdmscience.alphagenome.v1main.OntologyType\x12\n\n\x02id\x18\x02 \x01(\x03\"~\n\tBiosample\x12\x45\n\x04type\x18\x01 \x01(\x0e\x32\x37.google.gdm.gdmscience.alphagenome.v1main.BiosampleType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\x05stage\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_stage\"\x8b\x02\n\x12GeneScorerMetadata\x12\x0f\n\x07gene_id\x18\x01 \x01(\t\x12\x11\n\x04name\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x45\n\x06strand\x18\x02 \x01(\x0e\x32\x30.google.gdm.gdmscience.alphagenome.v1main.StrandH\x01\x88\x01\x01\x12\x11\n\x04type\x18\x04 \x01(\tH\x02\x88\x01\x01\x12\x1b\n\x0ejunction_start\x18\x05 \x01(\x03H\x03\x88\x01\x01\x12\x19\n\x0cjunction_end\x18\x06 \x01(\x03H\x04\x88\x01\x01\x42\x07\n\x05_nameB\t\n\x07_strandB\x07\n\x05_typeB\x11\n\x0f_junction_startB\x0f\n\r_junction_end\"\xa7\x05\n\rTrackMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12@\n\x06strand\x18\x02 \x01(\x0e\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Strand\x12M\n\rontology_term\x18\x03 \x01(\x0b\x32\x36.google.gdm.gdmscience.alphagenome.v1main.OntologyTerm\x12\x46\n\tbiosample\x18\x04 \x01(\x0b\x32\x33.google.gdm.gdmscience.alphagenome.v1main.Biosample\x12\x12\n\x05\x61ssay\x18\x05 \x01(\tH\x00\x88\x01\x01\x12\x1e\n\x11histone_mark_code\x18\t \x01(\tH\x01\x88\x01\x01\x12&\n\x19transcription_factor_code\x18\n \x01(\tH\x02\x88\x01\x01\x12\x18\n\x0bgtex_tissue\x18\x08 \x01(\tH\x03\x88\x01\x01\x12\x18\n\x0b\x64\x61ta_source\x18\x0b \x01(\tH\x04\x88\x01\x01\x12K\n\tendedness\x18\x0c \x01(\x0e\x32\x33.google.gdm.gdmscience.alphagenome.v1main.EndednessH\x05\x88\x01\x01\x12!\n\x14genetically_modified\x18\r \x01(\x08H\x06\x88\x01\x01\x12\x19\n\x0cnonzero_mean\x18\x0e \x01(\x02H\x07\x88\x01\x01\x42\x08\n\x06_assayB\x14\n\x12_histone_mark_codeB\x1c\n\x1a_transcription_factor_codeB\x0e\n\x0c_gtex_tissueB\x0e\n\x0c_data_sourceB\x0c\n\n_endednessB\x17\n\x15_genetically_modifiedB\x0f\n\r_nonzero_mean\"[\n\x0eTracksMetadata\x12I\n\x08metadata\x18\x01 \x03(\x0b\x32\x37.google.gdm.gdmscience.alphagenome.v1main.TrackMetadata\"\xa9\x02\n\x10JunctionMetadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12M\n\rontology_term\x18\x02 \x01(\x0b\x32\x36.google.gdm.gdmscience.alphagenome.v1main.OntologyTerm\x12\x46\n\tbiosample\x18\x03 \x01(\x0b\x32\x33.google.gdm.gdmscience.alphagenome.v1main.Biosample\x12\x18\n\x0bgtex_tissue\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0b\x64\x61ta_source\x18\x05 \x01(\tH\x01\x88\x01\x01\x12\x12\n\x05\x61ssay\x18\x06 \x01(\tH\x02\x88\x01\x01\x42\x0e\n\x0c_gtex_tissueB\x0e\n\x0c_data_sourceB\x08\n\x06_assay\"a\n\x11JunctionsMetadata\x12L\n\x08metadata\x18\x01 \x03(\x0b\x32:.google.gdm.gdmscience.alphagenome.v1main.JunctionMetadata\"\x86\x02\n\tTrackData\x12@\n\x06values\x18\x01 \x01(\x0b\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Tensor\x12I\n\x08metadata\x18\x02 \x03(\x0b\x32\x37.google.gdm.gdmscience.alphagenome.v1main.TrackMetadata\x12\x17\n\nresolution\x18\x03 \x01(\x03H\x00\x88\x01\x01\x12\x44\n\x08interval\x18\x04 \x01(\x0b\x32\x32.google.gdm.gdmscience.alphagenome.v1main.IntervalB\r\n\x0b_resolution\"\xab\x02\n\x0cJunctionData\x12@\n\x06values\x18\x01 \x01(\x0b\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Tensor\x12L\n\x08metadata\x18\x02 \x03(\x0b\x32:.google.gdm.gdmscience.alphagenome.v1main.JunctionMetadata\x12\x45\n\tjunctions\x18\x03 \x03(\x0b\x32\x32.google.gdm.gdmscience.alphagenome.v1main.Interval\x12\x44\n\x08interval\x18\x04 \x01(\x0b\x32\x32.google.gdm.gdmscience.alphagenome.v1main.Interval\"\xfe\x01\n\x10IntervalMetadata\x12\x44\n\x08interval\x18\x01 \x01(\x0b\x32\x32.google.gdm.gdmscience.alphagenome.v1main.Interval\x12O\n\x0etrack_metadata\x18\x02 \x03(\x0b\x32\x37.google.gdm.gdmscience.alphagenome.v1main.TrackMetadata\x12S\n\rgene_metadata\x18\x03 \x03(\x0b\x32<.google.gdm.gdmscience.alphagenome.v1main.GeneScorerMetadata\"\x9e\x01\n\x0cIntervalData\x12@\n\x06values\x18\x01 \x01(\x0b\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Tensor\x12L\n\x08metadata\x18\x02 \x01(\x0b\x32:.google.gdm.gdmscience.alphagenome.v1main.IntervalMetadata\"\xfb\x01\n\x0fVariantMetadata\x12\x42\n\x07variant\x18\x01 \x01(\x0b\x32\x31.google.gdm.gdmscience.alphagenome.v1main.Variant\x12O\n\x0etrack_metadata\x18\x02 \x03(\x0b\x32\x37.google.gdm.gdmscience.alphagenome.v1main.TrackMetadata\x12S\n\rgene_metadata\x18\x03 \x03(\x0b\x32<.google.gdm.gdmscience.alphagenome.v1main.GeneScorerMetadata\"\x9c\x01\n\x0bVariantData\x12@\n\x06values\x18\x01 \x01(\x0b\x32\x30.google.gdm.gdmscience.alphagenome.v1main.Tensor\x12K\n\x08metadata\x18\x02 \x01(\x0b\x32\x39.google.gdm.gdmscience.alphagenome.v1main.VariantMetadata\"\xbc\x02\n\x06Output\x12I\n\x0boutput_type\x18\x01 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputType\x12I\n\ntrack_data\x18\x02 \x01(\x0b\x32\x33.google.gdm.gdmscience.alphagenome.v1main.TrackDataH\x00\x12@\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x30.google.gdm.gdmscience.alphagenome.v1main.TensorH\x00\x12O\n\rjunction_data\x18\x04 \x01(\x0b\x32\x36.google.gdm.gdmscience.alphagenome.v1main.JunctionDataH\x00\x42\t\n\x07payload\"d\n\x13ScoreIntervalOutput\x12M\n\rinterval_data\x18\x01 \x01(\x0b\x32\x36.google.gdm.gdmscience.alphagenome.v1main.IntervalData\"a\n\x12ScoreVariantOutput\x12K\n\x0cvariant_data\x18\x01 \x01(\x0b\x32\x35.google.gdm.gdmscience.alphagenome.v1main.VariantData\"\xd4\x01\n\x16GeneMaskIntervalScorer\x12N\n\x10requested_output\x18\x01 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputType\x12\r\n\x05width\x18\x02 \x01(\x03\x12[\n\x10\x61ggregation_type\x18\x03 \x01(\x0e\x32\x41.google.gdm.gdmscience.alphagenome.v1main.IntervalAggregationType\"q\n\x0eIntervalScorer\x12U\n\tgene_mask\x18\x01 \x01(\x0b\x32@.google.gdm.gdmscience.alphagenome.v1main.GeneMaskIntervalScorerH\x00\x42\x08\n\x06scorer\"\xd5\x01\n\x10\x43\x65nterMaskScorer\x12\x12\n\x05width\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12S\n\x10\x61ggregation_type\x18\x02 \x01(\x0e\x32\x39.google.gdm.gdmscience.alphagenome.v1main.AggregationType\x12N\n\x10requested_output\x18\x03 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputTypeB\x08\n\x06_width\"c\n\x11GeneMaskLFCScorer\x12N\n\x10requested_output\x18\x01 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputType\"f\n\x14GeneMaskActiveScorer\x12N\n\x10requested_output\x18\x01 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputType\"\x86\x01\n\x16GeneMaskSplicingScorer\x12\x12\n\x05width\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12N\n\x10requested_output\x18\x02 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputTypeB\x08\n\x06_width\"\x17\n\x15PolyadenylationScorer\"\x16\n\x14SpliceJunctionScorer\"\x12\n\x10\x43ontactMapScorer\"\xfb\x04\n\rVariantScorer\x12Q\n\x0b\x63\x65nter_mask\x18\x01 \x01(\x0b\x32:.google.gdm.gdmscience.alphagenome.v1main.CenterMaskScorerH\x00\x12P\n\tgene_mask\x18\x02 \x01(\x0b\x32;.google.gdm.gdmscience.alphagenome.v1main.GeneMaskLFCScorerH\x00\x12^\n\x12gene_mask_splicing\x18\x03 \x01(\x0b\x32@.google.gdm.gdmscience.alphagenome.v1main.GeneMaskSplicingScorerH\x00\x12Q\n\x06pa_qtl\x18\x04 \x01(\x0b\x32?.google.gdm.gdmscience.alphagenome.v1main.PolyadenylationScorerH\x00\x12Y\n\x0fsplice_junction\x18\x05 \x01(\x0b\x32>.google.gdm.gdmscience.alphagenome.v1main.SpliceJunctionScorerH\x00\x12Q\n\x0b\x63ontact_map\x18\x06 \x01(\x0b\x32:.google.gdm.gdmscience.alphagenome.v1main.ContactMapScorerH\x00\x12Z\n\x10gene_mask_active\x18\x07 \x01(\x0b\x32>.google.gdm.gdmscience.alphagenome.v1main.GeneMaskActiveScorerH\x00\x42\x08\n\x06scorer\"\x84\x02\n\x0eOutputMetadata\x12I\n\x0boutput_type\x18\x01 \x01(\x0e\x32\x34.google.gdm.gdmscience.alphagenome.v1main.OutputType\x12J\n\x06tracks\x18\x02 \x01(\x0b\x32\x38.google.gdm.gdmscience.alphagenome.v1main.TracksMetadataH\x00\x12P\n\tjunctions\x18\x03 \x01(\x0b\x32;.google.gdm.gdmscience.alphagenome.v1main.JunctionsMetadataH\x00\x42\t\n\x07payload*a\n\x06Strand\x12\x16\n\x12STRAND_UNSPECIFIED\x10\x00\x12\x13\n\x0fSTRAND_POSITIVE\x10\x01\x12\x13\n\x0fSTRAND_NEGATIVE\x10\x02\x12\x15\n\x11STRAND_UNSTRANDED\x10\x03*\xa2\x01\n\x0cOntologyType\x12\x1d\n\x19ONTOLOGY_TYPE_UNSPECIFIED\x10\x00\x12\x15\n\x11ONTOLOGY_TYPE_CLO\x10\x01\x12\x18\n\x14ONTOLOGY_TYPE_UBERON\x10\x02\x12\x14\n\x10ONTOLOGY_TYPE_CL\x10\x03\x12\x15\n\x11ONTOLOGY_TYPE_EFO\x10\x04\x12\x15\n\x11ONTOLOGY_TYPE_NTR\x10\x05*\xfd\x01\n\rBiosampleType\x12\x1e\n\x1a\x42IOSAMPLE_TYPE_UNSPECIFIED\x10\x00\x12\x1f\n\x1b\x42IOSAMPLE_TYPE_PRIMARY_CELL\x10\x01\x12\x30\n,BIOSAMPLE_TYPE_IN_VITRO_DIFFERENTIATED_CELLS\x10\x02\x12\x1c\n\x18\x42IOSAMPLE_TYPE_CELL_LINE\x10\x03\x12\x19\n\x15\x42IOSAMPLE_TYPE_TISSUE\x10\x04\x12#\n\x1f\x42IOSAMPLE_TYPE_TECHNICAL_SAMPLE\x10\x05\x12\x1b\n\x17\x42IOSAMPLE_TYPE_ORGANOID\x10\x06*\xd5\x02\n\nOutputType\x12\x1b\n\x17OUTPUT_TYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10OUTPUT_TYPE_ATAC\x10\x01\x12\x14\n\x10OUTPUT_TYPE_CAGE\x10\x02\x12\x15\n\x11OUTPUT_TYPE_DNASE\x10\x03\x12\x17\n\x13OUTPUT_TYPE_RNA_SEQ\x10\x04\x12\x1c\n\x18OUTPUT_TYPE_CHIP_HISTONE\x10\x05\x12\x17\n\x13OUTPUT_TYPE_CHIP_TF\x10\x06\x12\x1c\n\x18OUTPUT_TYPE_SPLICE_SITES\x10\x07\x12!\n\x1dOUTPUT_TYPE_SPLICE_SITE_USAGE\x10\x08\x12 \n\x1cOUTPUT_TYPE_SPLICE_JUNCTIONS\x10\t\x12\x1c\n\x18OUTPUT_TYPE_CONTACT_MAPS\x10\x0b\x12\x16\n\x12OUTPUT_TYPE_PROCAP\x10\x0c*\\\n\x08Organism\x12\x18\n\x14ORGANISM_UNSPECIFIED\x10\x00\x12\x1a\n\x15ORGANISM_HOMO_SAPIENS\x10\x86K\x12\x1a\n\x15ORGANISM_MUS_MUSCULUS\x10\xeaN*\x8b\x01\n\x17IntervalAggregationType\x12)\n%INTERVAL_AGGREGATION_TYPE_UNSPECIFIED\x10\x00\x12\"\n\x1eINTERVAL_AGGREGATION_TYPE_MEAN\x10\x01\x12!\n\x1dINTERVAL_AGGREGATION_TYPE_SUM\x10\x02*\xbf\x02\n\x0f\x41ggregationType\x12 \n\x1c\x41GGREGATION_TYPE_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x41GGREGATION_TYPE_DIFF_MEAN\x10\x01\x12\x1d\n\x19\x41GGREGATION_TYPE_DIFF_SUM\x10\x02\x12\"\n\x1e\x41GGREGATION_TYPE_DIFF_SUM_LOG2\x10\x03\x12\x1c\n\x18\x41GGREGATION_TYPE_L2_DIFF\x10\x04\x12\"\n\x1e\x41GGREGATION_TYPE_L2_DIFF_LOG1P\x10\x08\x12\"\n\x1e\x41GGREGATION_TYPE_DIFF_LOG2_SUM\x10\x05\x12 \n\x1c\x41GGREGATION_TYPE_ACTIVE_MEAN\x10\x06\x12\x1f\n\x1b\x41GGREGATION_TYPE_ACTIVE_SUM\x10\x07*R\n\tEndedness\x12\x19\n\x15\x45NDEDNESS_UNSPECIFIED\x10\x00\x12\x14\n\x10\x45NDEDNESS_SINGLE\x10\x01\x12\x14\n\x10\x45NDEDNESS_PAIRED\x10\x02\x42\x94\x01\n,com.google.gdm.gdmscience.alphagenome.v1mainB\rDnaModelProtoP\x01ZSgoogle.golang.org/genproto/googleapis/gdm/gdmscience/alphagenome/v1main;alphagenomeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'alphagenome.protos.dna_model_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.google.gdm.gdmscience.alphagenome.v1mainB\rDnaModelProtoP\001ZSgoogle.golang.org/genproto/googleapis/gdm/gdmscience/alphagenome/v1main;alphagenome'
  _globals['_STRAND']._serialized_start=5793
  _globals['_STRAND']._serialized_end=5890
  _globals['_ONTOLOGYTYPE']._serialized_start=5893
  _globals['_ONTOLOGYTYPE']._serialized_end=6055
  _globals['_BIOSAMPLETYPE']._serialized_start=6058
  _globals['_BIOSAMPLETYPE']._serialized_end=6311
  _globals['_OUTPUTTYPE']._serialized_start=6314
  _globals['_OUTPUTTYPE']._serialized_end=6655
  _globals['_ORGANISM']._serialized_start=6657
  _globals['_ORGANISM']._serialized_end=6749
  _globals['_INTERVALAGGREGATIONTYPE']._serialized_start=6752
  _globals['_INTERVALAGGREGATIONTYPE']._serialized_end=6891
  _globals['_AGGREGATIONTYPE']._serialized_start=6894
  _globals['_AGGREGATIONTYPE']._serialized_end=7213
  _globals['_ENDEDNESS']._serialized_start=7215
  _globals['_ENDEDNESS']._serialized_end=7297
  _globals['_INTERVAL']._serialized_start=113
  _globals['_INTERVAL']._serialized_end=237
  _globals['_VARIANT']._serialized_start=239
  _globals['_VARIANT']._serialized_end=336
  _globals['_ONTOLOGYTERM']._serialized_start=338
  _globals['_ONTOLOGYTERM']._serialized_end=443
  _globals['_BIOSAMPLE']._serialized_start=445
  _globals['_BIOSAMPLE']._serialized_end=571
  _globals['_GENESCORERMETADATA']._serialized_start=574
  _globals['_GENESCORERMETADATA']._serialized_end=841
  _globals['_TRACKMETADATA']._serialized_start=844
  _globals['_TRACKMETADATA']._serialized_end=1523
  _globals['_TRACKSMETADATA']._serialized_start=1525
  _globals['_TRACKSMETADATA']._serialized_end=1616
  _globals['_JUNCTIONMETADATA']._serialized_start=1619
  _globals['_JUNCTIONMETADATA']._serialized_end=1916
  _globals['_JUNCTIONSMETADATA']._serialized_start=1918
  _globals['_JUNCTIONSMETADATA']._serialized_end=2015
  _globals['_TRACKDATA']._serialized_start=2018
  _globals['_TRACKDATA']._serialized_end=2280
  _globals['_JUNCTIONDATA']._serialized_start=2283
  _globals['_JUNCTIONDATA']._serialized_end=2582
  _globals['_INTERVALMETADATA']._serialized_start=2585
  _globals['_INTERVALMETADATA']._serialized_end=2839
  _globals['_INTERVALDATA']._serialized_start=2842
  _globals['_INTERVALDATA']._serialized_end=3000
  _globals['_VARIANTMETADATA']._serialized_start=3003
  _globals['_VARIANTMETADATA']._serialized_end=3254
  _globals['_VARIANTDATA']._serialized_start=3257
  _globals['_VARIANTDATA']._serialized_end=3413
  _globals['_OUTPUT']._serialized_start=3416
  _globals['_OUTPUT']._serialized_end=3732
  _globals['_SCOREINTERVALOUTPUT']._serialized_start=3734
  _globals['_SCOREINTERVALOUTPUT']._serialized_end=3834
  _globals['_SCOREVARIANTOUTPUT']._serialized_start=3836
  _globals['_SCOREVARIANTOUTPUT']._serialized_end=3933
  _globals['_GENEMASKINTERVALSCORER']._serialized_start=3936
  _globals['_GENEMASKINTERVALSCORER']._serialized_end=4148
  _globals['_INTERVALSCORER']._serialized_start=4150
  _globals['_INTERVALSCORER']._serialized_end=4263
  _globals['_CENTERMASKSCORER']._serialized_start=4266
  _globals['_CENTERMASKSCORER']._serialized_end=4479
  _globals['_GENEMASKLFCSCORER']._serialized_start=4481
  _globals['_GENEMASKLFCSCORER']._serialized_end=4580
  _globals['_GENEMASKACTIVESCORER']._serialized_start=4582
  _globals['_GENEMASKACTIVESCORER']._serialized_end=4684
  _globals['_GENEMASKSPLICINGSCORER']._serialized_start=4687
  _globals['_GENEMASKSPLICINGSCORER']._serialized_end=4821
  _globals['_POLYADENYLATIONSCORER']._serialized_start=4823
  _globals['_POLYADENYLATIONSCORER']._serialized_end=4846
  _globals['_SPLICEJUNCTIONSCORER']._serialized_start=4848
  _globals['_SPLICEJUNCTIONSCORER']._serialized_end=4870
  _globals['_CONTACTMAPSCORER']._serialized_start=4872
  _globals['_CONTACTMAPSCORER']._serialized_end=4890
  _globals['_VARIANTSCORER']._serialized_start=4893
  _globals['_VARIANTSCORER']._serialized_end=5528
  _globals['_OUTPUTMETADATA']._serialized_start=5531
  _globals['_OUTPUTMETADATA']._serialized_end=5791
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: alphagenome/protos/tensor.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'alphagenome/protos/tensor.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x61lphagenome/protos/tensor.proto\x12(google.gdm.gdmscience.alphagenome.v1main\"\xc8\x01\n\x06Tensor\x12\r\n\x05shape\x18\x01 \x03(\x05\x12\x45\n\tdata_type\x18\x02 \x01(\x0e\x32\x32.google.gdm.gdmscience.alphagenome.v1main.DataType\x12\x46\n\x05\x61rray\x18\x03 \x01(\x0b\x32\x35.google.gdm.gdmscience.alphagenome.v1main.TensorChunkH\x00\x12\x15\n\x0b\x63hunk_count\x18\x04 \x01(\x03H\x00\x42\t\n\x07payload\"p\n\x0bTensorChunk\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12S\n\x10\x63ompression_type\x18\x02 \x01(\x0e\x32\x39.google.gdm.gdmscience.alphagenome.v1main.CompressionType*\x95\x02\n\x08\x44\x61taType\x12\x19\n\x15\x44\x41TA_TYPE_UNSPECIFIED\x10\x00\x12\x16\n\x12\x44\x41TA_TYPE_BFLOAT16\x10\x01\x12\x15\n\x11\x44\x41TA_TYPE_FLOAT16\x10\x0b\x12\x15\n\x11\x44\x41TA_TYPE_FLOAT32\x10\x02\x12\x15\n\x11\x44\x41TA_TYPE_FLOAT64\x10\x03\x12\x12\n\x0e\x44\x41TA_TYPE_INT8\x10\x04\x12\x13\n\x0f\x44\x41TA_TYPE_INT32\x10\x05\x12\x13\n\x0f\x44\x41TA_TYPE_INT64\x10\x06\x12\x13\n\x0f\x44\x41TA_TYPE_UINT8\x10\x07\x12\x14\n\x10\x44\x41TA_TYPE_UINT32\x10\x08\x12\x14\n\x10\x44\x41TA_TYPE_UINT64\x10\t\x12\x12\n\x0e\x44\x41TA_TYPE_BOOL\x10\n*G\n\x0f\x43ompressionType\x12\x19\n\x15\x43OMPRESSION_TYPE_NONE\x10\x00\x12\x19\n\x15\x43OMPRESSION_TYPE_ZSTD\x10\x01\x42\x92\x01\n,com.google.gdm.gdmscience.alphagenome.v1mainB\x0bTensorProtoP\x01ZSgoogle.golang.org/genproto/googleapis/gdm/gdmscience/alphagenome/v1main;alphagenomeb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'alphagenome.protos.tensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n,com.google.gdm.gdmscience.alphagenome.v1mainB\013TensorProtoP\001ZSgoogle.golang.org/genproto/googleapis/gdm/gdmscience/alphagenome/v1main;alphagenome'
  _globals['_DATATYPE']._serialized_start=395
  _globals['_DATATYPE']._serialized_end=672
  _globals['_COMPRESSIONTYPE']._serialized_start=674
  _globals['_COMPRESSIONTYPE']._serialized_end=745
  _globals['_TENSOR']._serialized_start=78
  _globals['_TENSOR']._serialized_end=278
  _globals['_TENSORCHUNK']._serialized_start=280
  _globals['_TENSORCHUNK']._serialized_end=392
# @@protoc_insertion_point(module_scope)
